# Import models for the CLI registry
from watch.tasks.fusion.methods import *  # NOQA
from watch.tasks.fusion.datamodules.kwcoco_datamodule import KWCocoVideoDataModule
from watch.utils.lightning_ext.lightning_cli_ext import LightningCLI_Extension
from watch.utils import lightning_ext as pl_ext

import pytorch_lightning as pl
import ubelt as ub
from torch import optim
from typing import Any

import yaml
from jsonargparse import set_loader, set_dumper
# from pytorch_lightning.utilities.rank_zero import rank_zero_only


# Not very safe, but needed to parse tuples e.g. datamodule.dataset_stats
# TODO: yaml.SafeLoader + tuple parsing
def custom_yaml_load(stream):
    return yaml.load(stream, Loader=yaml.FullLoader)


set_loader('yaml_unsafe_for_tuples', custom_yaml_load)


def custom_yaml_dump(data):
    return yaml.dump(data, Dumper=yaml.Dumper)


set_dumper('yaml_unsafe_for_tuples', custom_yaml_dump)


class SmartTrainer(pl.Trainer):
    """
    Simple trainer subclass so we can ensure a print happens directly before
    the training loop. (so annoying that we can't reorder callbacks)
    """
    ...

    def _run_stage(self, *args, **kwargs):
        # All I want is to print this  directly before training starts.
        # Is that so hard to do?
        print(f'self.global_rank={self.global_rank}')
        if self.global_rank == 0:
            import rich
            dpath = self.logger.log_dir
            rich.print(f"Trainer log dpath:\n\n[link={dpath}]{dpath}[/link]\n")
        super()._run_stage(*args, **kwargs)


class TorchGlobals(pl.callbacks.Callback):
    """
    Callback to setup torch globals

    Args:
        float32_matmul_precision (str):
            can be 'medium', 'high', 'default', or 'auto'.
            The 'default' value does not change any setting.
            The 'auto' value defaults to 'medium' if the training devices have
                ampere cores.
    """

    def __init__(self, float32_matmul_precision='default'):
        self.float32_matmul_precision = float32_matmul_precision

    def setup(self, trainer, pl_module, stage):
        import torch
        float32_matmul_precision = self.float32_matmul_precision
        if float32_matmul_precision == 'default':
            float32_matmul_precision = None
        elif float32_matmul_precision == 'auto':
            # Detect if we have Ampere tensor cores
            # Ampere (V8) and later leverage tensor cores, where medium
            # float32_matmul_precision becomes useful
            if torch.cuda.is_available():
                device_versions = [torch.cuda.get_device_capability(device_id)[0]
                                   for device_id in trainer.device_ids]
                if all(v >= 8 for v in device_versions):
                    float32_matmul_precision = 'medium'
                else:
                    float32_matmul_precision = None
            else:
                float32_matmul_precision = None
        if float32_matmul_precision is not None:
            torch.set_float32_matmul_precision(float32_matmul_precision)


class WeightInitializer(pl.callbacks.Callback):
    """
    Netowrk weight initializer with support for partial weight loading.
    """

    def __init__(self, init='noop', association='embedding'):
        self.init = init
        self.association = association
        self._did_once = False

    def setup(self, trainer, pl_module, stage):
        if self._did_once:
            # Only weight initialize once, on whatever stage happens first.
            return
        self._did_once = True
        if self.init != 'noop':
            from watch.tasks.fusion.fit import coerce_initializer
            from watch.utils import util_pattern
            initializer = coerce_initializer(self.init)

            model = pl_module

            # Hack to always preserve specific values
            # (TODO add to torch_liberator as an option, allow to be configured
            # as argument to this class, or allow the model to set some property
            # that says which weights should not be touched.)
            print('Initializing weights')
            old_state = model.state_dict()
            ignore_pattern = util_pattern.MultiPattern.coerce(['*tokenizers*.0.mean', '*tokenizers*.0.std'])
            ignore_keys = [key for key in old_state.keys() if ignore_pattern.match(key)]
            print('Finding keys to not initializer')
            to_preserve = ub.udict(old_state).subdict(ignore_keys).map_values(lambda v: v.clone())

            initializer.association = self.association
            info = initializer.forward(model)  # NOQA
            if info:
                mapping = info.get('mapping', None)
                unset = info.get('self_unset', None)
                unused = info.get('self_unused', None)
                print('mapping = {}'.format(ub.urepr(mapping, nl=1)))
                print(f'unused={unused}')
                print(f'unset={unset}')

            print('Finalize initialization')
            updated = model.state_dict() | to_preserve
            model.load_state_dict(updated)


class SmartLightningCLI(LightningCLI_Extension):

    @staticmethod
    def configure_optimizers(
        lightning_module: pl.LightningModule, optimizer: optim.Optimizer, lr_scheduler=None
    ) -> Any:
        """Override to customize the :meth:`~pytorch_lightning.core.module.LightningModule.configure_optimizers`
        method.

        Args:
            lightning_module: A reference to the model.
            optimizer: The optimizer.
            lr_scheduler: The learning rate scheduler (if used).
        """

        if lr_scheduler is None:
            return optimizer

        if isinstance(lr_scheduler, pl.cli.ReduceLROnPlateau):
            return {
                "optimizer": optimizer,
                "lr_scheduler": {"scheduler": lr_scheduler, "monitor": lr_scheduler.monitor},
            }

        if isinstance(lr_scheduler, optim.lr_scheduler.OneCycleLR):
            return {
                "optimizer": optimizer,
                "lr_scheduler": {"scheduler": lr_scheduler, 'interval': 'step'},
            }

        return [optimizer], [lr_scheduler]

    def add_arguments_to_parser(self, parser):

        # TODO: separate final_package dir and fpath for more configuration
        # pl_ext.callbacks.Packager(package_fpath=args.package_fpath),
        parser.add_lightning_class_args(pl_ext.callbacks.Packager, "packager")

        parser.add_lightning_class_args(WeightInitializer, "initializer")

        parser.add_lightning_class_args(TorchGlobals, "torch_globals")

        parser.add_lightning_class_args(pl_ext.callbacks.BatchPlotter, "batch_plotter")
        # pl_ext.callbacks.BatchPlotter(  # Fixme: disabled for multi-gpu training with deepspeed
        #     num_draw=2,  # args.num_draw,
        #     draw_interval="5min",  # args.draw_interval
        # ),

        # parser.set_defaults({"packager.package_fpath": "???"}) # "$DEFAULT_ROOT_DIR"/final_package.pt
        parser.link_arguments(
            "trainer.default_root_dir",
            "packager.package_fpath",
            compute_fn=lambda root: None if root is None else str(ub.Path(root) / "final_package.pt")
            # apply_on="instantiate",
        )

        def data_value_getter(key):
            # Hack to call setup on the datamodule before linking args
            def get_value(data):
                if not data.did_setup:
                    data.setup('fit')
                return getattr(data, key)
            return get_value

        # pass dataset stats to model after initialization datamodule
        parser.link_arguments(
            "data",
            "model.init_args.dataset_stats",
            compute_fn=data_value_getter('dataset_stats'),
            apply_on="instantiate")
        parser.link_arguments(
            "data",
            "model.init_args.classes",
            compute_fn=data_value_getter('classes'),
            apply_on="instantiate")

        super().add_arguments_to_parser(parser)


def make_cli(config=None):
    """
    Note:
        Currently, creating the CLI will invoke it. We could modify this
        function to have the option to not invoke by specifying ``run=False``
        to :class:`LightningCLI`, but for some reason that changes the expected
        form of the config (you must specify subcommand if run=True but must
        not if run=False). We need to understand exactly what's going on there
        before we expose a way to set run=False.
    """

    if isinstance(config, str):
        try:
            if len(config) > 200:
                raise Exception
            if ub.Path(config).exists():
                config = config.read_text()
        except Exception:
            ...

        def nested_to_jsonnest(nested):
            config = {}
            for p, v in ub.IndexableWalker(nested):
                if not isinstance(v, (dict, list)):
                    k = '.'.join(list(map(str, p)))
                    config[k] = v
            return config
        from watch.utils import util_yaml
        print('Passing string-based config:')
        print(ub.highlight_code(config, 'yaml'))
        nested = util_yaml.Yaml.loads(config, backend='pyyaml')
        # print('nested = {}'.format(ub.urepr(nested, nl=1)))
        config = nested_to_jsonnest(nested)
        # print('config = {}'.format(ub.urepr(config, nl=1)))

    clikw = {'run': True}
    if config is not None:
        # overload the argument parsing with a programatic config
        clikw['args'] = config
        # Note: we may not need manual mode by setting run to False once we
        # have a deeper understanding of how lightning CLI works.
        # clikw['run'] = False

    default_callbacks = [
        pl.callbacks.RichProgressBar(),
        # pl.callbacks.LearningRateMonitor(logging_interval='step', log_momentum=True),

        pl.callbacks.LearningRateMonitor(logging_interval='epoch', log_momentum=True),
        # pl.callbacks.ModelCheckpoint(monitor='train_loss', mode='min', save_top_k=1),

        # leaving always on breaks when correspinding metric isnt
        # tracked because loss_weight==0
        # FIXME: can we conditionally apply these if they make sense?
        # or can we make them robust to the case where the key isn't logged?
        # pl.callbacks.ModelCheckpoint(
        #     monitor='val_change_f1', mode='max', save_top_k=4),
        # pl.callbacks.ModelCheckpoint(
        #     monitor='val_saliency_f1', mode='max', save_top_k=4),
        # pl.callbacks.ModelCheckpoint(
        #     monitor='val_class_f1_micro', mode='max', save_top_k=4),
        # pl.callbacks.ModelCheckpoint(
        #     monitor='val_class_f1_macro', mode='max', save_top_k=4),
    ]
    try:
        # There has to be a tool with less dependencies the matplotlib
        # auto-plotters can hook into.
        import tensorboard  # NOQA
    except ImportError:
        print('warning: tensorboard not available')
    else:
        # Only use tensorboard if we have it.
        default_callbacks.append(pl_ext.callbacks.TensorboardPlotter())

    cli = SmartLightningCLI(
        model_class=pl.LightningModule,  # TODO: factor out common components of the two models and put them in base class models inherit from
        datamodule_class=KWCocoVideoDataModule,
        trainer_class=SmartTrainer,
        subclass_mode_model=True,

        # save_config_overwrite=True,
        save_config_kwargs={
            'overwrite': True,
        },

        # subclass_mode_data=True,
        parser_kwargs=dict(
            parser_mode='yaml_unsafe_for_tuples',
            error_handler=None,
        ),
        trainer_defaults=dict(
            # The following works, but it might be better to move some of these callbacks into the cli
            # (https://pytorch-lightning.readthedocs.io/en/latest/cli/lightning_cli_expert.html#configure-forced-callbacks)
            # Another option is to have a base_config.yaml that includes these, which would make them fully configurable
            # without modifying source code.
            # TODO: find good way to reenable profiling, but not by default
            # profiler=pl.profilers.AdvancedProfiler(dirpath=".", filename="perf_logs"),

            callbacks=default_callbacks,
        ),
        **clikw,
    )
    return cli


def main(config=None):
    """
    Args:
        config (None | Dict):
            if specified disables sys.argv usage and executes a training run
            with the specified config.

    CommandLine:
        xdoctest -m watch.tasks.fusion.fit_lightning main:0

    Ignore:
        ...

        # export stats
        from watch.utils.lightning_ext.monkeypatches import disable_lightning_hardware_warnings
        from watch.tasks.fusion.fit_lightning import *  # NOQA
        from watch.utils.util_yaml import Yaml
        disable_lightning_hardware_warnings()

        def export_dataset_stats(cli):
            input_stats = cli.trainer.datamodule.dataset_stats['input_stats']
            rows = []
            for sensorchan, stats in input_stats.items():
                mean = list(map(float, stats['mean'].ravel().tolist()))
                std = list(map(float, stats['std'].ravel().tolist()))
                row = {
                    'sensor': sensorchan[0],
                    'channels': sensorchan[1],
                    'mean': Yaml.InlineList(mean),
                    'std': Yaml.InlineList(std),
                }
                rows.append(row)
            from watch.utils import util_yaml
            import ruamel.yaml
            import io
            file = io.StringIO()
            ruamel.yaml.round_trip_dump(rows, file, Dumper=ruamel.yaml.RoundTripDumper)
            print(file.getvalue())
        dataset_stats = ub.codeblock(
            '''
            - sensor: '*'
              channels: r|g|b
              mean: [87.572401, 87.572401, 87.572401]
              std: [99.449996, 99.449996, 99.449996]
            ''')
        dpath = ub.Path.appdir('watch/tests/test_fusion_fit/demo_main_noop').delete().ensuredir()
        config = {
            'subcommand': 'fit',
            'fit.model': 'watch.tasks.fusion.methods.noop_model.NoopModel',
            'fit.trainer.default_root_dir': dpath,
            'fit.data.train_dataset': 'special:vidshapes4-frames9-gsize32',
            'fit.data.vali_dataset': 'special:vidshapes1-frames9-gsize32',
            'fit.data.window_dims': 32,
            'fit.data.dataset_stats': dataset_stats,
            'fit.trainer.max_steps': 2,
            'fit.trainer.num_sanity_val_steps': 0,
        }
        cli = main(config=config)


    Example:
        >>> from watch.utils.lightning_ext.monkeypatches import disable_lightning_hardware_warnings
        >>> from watch.tasks.fusion.fit_lightning import *  # NOQA
        >>> disable_lightning_hardware_warnings()
        >>> dpath = ub.Path.appdir('watch/tests/test_fusion_fit/demo_main_noop').delete().ensuredir()
        >>> config = {
        >>>     'subcommand': 'fit',
        >>>     'fit.model': 'watch.tasks.fusion.methods.noop_model.NoopModel',
        >>>     'fit.trainer.default_root_dir': dpath,
        >>>     'fit.data.train_dataset': 'special:vidshapes4-frames9-gsize32',
        >>>     'fit.data.vali_dataset': 'special:vidshapes1-frames9-gsize32',
        >>>     'fit.data.chip_dims': 32,
        >>>     'fit.trainer.accelerator': 'cpu',
        >>>     'fit.trainer.devices': 1,
        >>>     'fit.trainer.max_steps': 2,
        >>>     'fit.trainer.num_sanity_val_steps': 0,
        >>> }
        >>> cli = main(config=config)

    Example:
        >>> from watch.utils.lightning_ext.monkeypatches import disable_lightning_hardware_warnings
        >>> from watch.tasks.fusion.fit_lightning import *  # NOQA
        >>> disable_lightning_hardware_warnings()
        >>> dpath = ub.Path.appdir('watch/tests/test_fusion_fit/demo_main_heterogeneous').delete().ensuredir()
        >>> config = {
        >>>     # 'model': 'watch.tasks.fusion.methods.MultimodalTransformer',
        >>>     #'model': 'watch.tasks.fusion.methods.UNetBaseline',
        >>>     'subcommand': 'fit',
        >>>     'fit.model.class_path': 'watch.tasks.fusion.methods.heterogeneous.HeterogeneousModel',
        >>>     'fit.optimizer.class_path': 'torch.optim.SGD',
        >>>     'fit.optimizer.init_args.lr': 1e-3,
        >>>     'fit.trainer.default_root_dir': dpath,
        >>>     'fit.data.train_dataset': 'special:vidshapes8-gsize64-frames9-speed0.5-multispectral',
        >>>     'fit.data.vali_dataset': 'special:vidshapes4-gsize64-frames9-speed0.5-multispectral',
        >>>     'fit.data.chip_dims': 64,
        >>>     'fit.trainer.accelerator': 'cpu',
        >>>     'fit.trainer.devices': 1,
        >>>     'fit.trainer.max_steps': 2,
        >>>     'fit.trainer.num_sanity_val_steps': 0,
        >>> }
        >>> main(config=config)
    """
    cli = make_cli(config)
    return cli


if __name__ == "__main__":
    r"""
    CommandLine:
        python -m watch.tasks.fusion.fit_lightning fit --help

        python -m watch.tasks.fusion.fit_lightning fit \
                --model.help=MultimodalTransformer

        python -m watch.tasks.fusion.fit_lightning fit \
                --model.help=NoopModel

        python -m watch.tasks.fusion.fit_lightning fit \
            --data.train_dataset=special:vidshapes8-frames9-speed0.5 \
            --data.window_dims=64 \
            --data.workers=4 \
            --trainer.accelerator=gpu \
            --trainer.strategy=ddp \
            --trainer.devices=0,1 \
            --data.batch_size=4 \
            --model.class_path=HeterogeneousModel \
            --optimizer.class_path=torch.optim.Adam \
            --trainer.default_root_dir ./demo_train

        python -m watch.tasks.fusion.fit_lightning fit --config="
            data:
                train_dataset: special:vidshapes8-frames9-speed0.5
                window_dims: 64
                workers: 4
                batch_size: 4
                normalize_inputs:
                    input_stats:
                        - sensor: '*'
                          channels: r|g|b
                          video: video1
                          mean: [87.572401, 87.572401, 87.572401]
                          std: [99.449996, 99.449996, 99.449996]
            trainer:
                accelerator: gpu
                strategy: ddp
                devices: 0,1
            model:
                class_path: HeterogeneousModel
            optimizer:
                class_path: torch.optim.Adam
            trainer:
                default_root_dir: ./demo_train
        "

        # Note: setting fast_dev_run seems to disable directory output.

        python -m watch.tasks.fusion.fit_lightning fit \
            --data.train_dataset=special:vidshapes8-frames9-speed0.5-multispectral \
            --trainer.accelerator=gpu \
            --trainer.devices=0, \
            --trainer.precision=16 \
            --trainer.fast_dev_run=5 \
            --model=NoopModel
    """
    main()



import numpy as np
from torch.utils.data import Dataset

from watch.tasks.rutgers_material_seg_v2.utils.util_misc import sigmoid


class BaseDataset(Dataset):
    def __init__(self, kwcoco_path, split, image_scale_mode='direct', seed_num=0, sensor='S2'):
        super().__init__()

        self.split = split
        self.sensor = sensor
        self.seed_num = seed_num
        self.kwcoco_path = kwcoco_path
        self.image_scale_mode = image_scale_mode

        # Get number of channels of input data.
        self.n_channels = self.get_n_channels()

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        raise NotImplementedError

    def get_n_channels(self):
        raise NotImplementedError

    def _scale_image(self, image, scale_mode='direct'):
        """Scales the pixel values of image.

        Args:
            image (np.array): A numpy array of shape [channels, height, width].
            scale_mode (str, optional): Parameter to determine how to scale image. Defaults to 'direct'.
        """
        def __direct_scale(image):
            """Map image values from whatever range to [0,1].

            Args:
                image (np.array): A np.array of any shape.

            Raises:
                NotImplementedError: Throws error when dtype conversion is not set.

            Returns:
                np.array: A np.arrary of any shape but with values between [0, 1].
            """
            if image.dtype == 'uint16':
                scaled_image = image / 2**16
            elif image.dtype == 'int16':
                scaled_image = image / 2**16
            elif image.dtype == 'float32':
                scaled_image = image
            else:
                raise NotImplementedError(f'method not implemented for scaling "{image.dtype}" type to [0,1] range.')

            # Make sure image is bounded between 0 and 1.
            scaled_image = np.clip(scaled_image, 0, 1)

            return scaled_image

        if scale_mode == 'direct':
            scaled_image = __direct_scale(image)
        elif scale_mode == 'sigmoid':
            # Idea taken from Dynamic World paper.

            # Apply log to values in image.
            log_image = np.log(image)

            # Apply sigmoid to map to [0,1] range.
            scaled_image = sigmoid(log_image)

        elif scale_mode == 'unit':
            image = __direct_scale(image)
            scaled_image = image / (np.linalg.norm(image, axis=0, ord=1) + 1e-4)

        else:
            raise NotImplementedError(f'Scale mode "{scale_mode}" not implemented.')

        return scaled_image

    def _load_norm_image(self, image_id, crop_slice=None, channels=None):
        """Load an image from kwcoco file and normalize to float.

        Args:
            image_id (int): Image ID (aka gid) to use as reference in kwcoco dataset.
            crop_slice (list, optional): Crop image to this rectangle, [h0, w0, hE, wE]. Defaults to None.
            channels (str, optional): Language specifying which channels to load from an image. Defaults to None.

        Returns:
            np.array: A float array of shape [channels, height, width].
        """
        if channels is None:
            channels = self.channels

        img_delayed = self.coco_dset.delayed_load(image_id, channels=channels)

        # Crop image (optional).
        if crop_slice is not None:
            h0, w0, dh, dw = crop_slice
            hE, wE = h0 + dh, w0 + dw
            img_delayed = img_delayed.crop((slice(h0, hE), slice(w0, wE)))

        # Finalize image loading.
        image = img_delayed.finalize()  # [height, width, channels]

        image = image.transpose(2, 0, 1)  # [channels, height, width]

        # Normalize image to [0, 1] range.
        image = self._scale_image(image, scale_mode=self.image_scale_mode)

        return image

    def _compute_dataset_stats(self, pixel_data):
        # Less sensitive to outliers
        mean = np.median(pixel_data, axis=-1)
        # mean = pixel_data.mean(axis=-1)
        std = pixel_data.std(axis=-1)

        return mean, std

    def _add_buffer_to_image(self, image, desired_height, desired_width, buffer_mode='constant', constant_value=0):
        """Extend image to desired size if smaller in resolution.
        Do nothing if the image is larger than the desired size.
        Args:
            image (np.array): A numpy array of shape [height, width] or [channels, height, width].
            desired_height (int): Desired height the image should be.
            desired_width (int): Desired width the image should be.
            buffer_mode (str, optional): Method mode for determining how to fill buffered regions. Defaults to 'constant'.
            constant_value (int, optional): For constant method, what value to assign to default canvas value. Defaults to 0.
        Raises:
            NotImplementedError: No method to handle images with number of dimensions other than 2 or 3.
            NotImplementedError: No method to handle images with number of dimensions other than 2 or 3.
        Returns:
            np.array: A numpy array of shape [desired_height, desired_width].
        """
        # Get image dimensions.
        n_dims = len(image.shape)
        if n_dims == 2:
            image_height, image_width = image.shape
        elif n_dims == 3:
            n_channels, image_height, image_width = image.shape
        else:
            raise NotImplementedError(f'Cannot add buffer to image with "{n_dims}" dimensions.')

        # Check if image is smaller than desired resolution.
        if (image_height >= desired_height) and (image_width >= desired_width):
            return image, np.ones([image_height, image_width], dtype=int)
        else:
            if buffer_mode == 'constant':
                # Create buffer canvas.
                buffer_mask = np.zeros([desired_height, desired_width], dtype=int)
                if n_dims == 2:
                    buffer_canvas = np.ones([desired_height, desired_width], dtype=image.dtype) * constant_value
                    buffer_canvas[:image_height, :image_width] = image

                elif n_dims == 3:
                    buffer_canvas = np.ones([n_channels, desired_height, desired_width],
                                            dtype=image.dtype) * constant_value
                    buffer_canvas[:, :image_height, :image_width] = image

                    buffer_mask = np.zeros([desired_height, desired_width], dtype=int)
                buffer_mask[:image_height, :image_width] = 1
                image = buffer_canvas
            else:
                raise NotImplementedError(f'No method to handle buffer mode of "{buffer_mode}"')

        return image, buffer_mask

    def _label_mask_manipulation(self, label_mask, label_polygon_raster=None):
        # TODO: Label smoothing
        ## Convert 1-hot encoding to label-distribution.

        # Label condensing (fuse material labels such as concrete and asphalt)

        # Attentuate label confidence (weight) by distnace to edge of polygon.
        ## Need to figure out algorithm for this.
        pass


if __name__ == '__main__':
    pass



import numpy as np
import ubelt as ub


def guess_missing_unixtimes(unixtimes):
    """
    Hueristic solution to fill in missing time values via interpolation /
    extrapolation.

    Example:
        >>> from watch.tasks.fusion.datamodules.temporal_sampling.utils import *  # NOQA
        >>> import ubelt as ub
        >>> cases = [
        >>>     np.array([np.nan, np.nan, np.nan, np.nan, np.nan]),
        >>>     np.array([np.nan, 20, 30, np.nan, np.nan]),
        >>>     np.array([0, np.nan, np.nan, np.nan, 10]),
        >>>     np.array([np.nan, np.nan, 9001, np.nan, np.nan]),
        >>>     np.array([1, 2, 3, 4, 5]),
        >>>     np.array([1, 2, np.nan, 4, 5]),
        >>> ]
        >>> for case_ in cases:
        >>>     unixtimes = case_
        >>>     print('case_ = {}'.format(ub.urepr(case_, nl=1)))
        >>>     guess = guess_missing_unixtimes(unixtimes)
        >>>     print('guess = {}'.format(ub.urepr(guess, nl=1)))
    """
    missing_date = np.isnan(unixtimes)
    num_missing = missing_date.sum()
    missing_any_dates = num_missing > 0

    if missing_any_dates:
        num_have = len(unixtimes) - num_missing
        have_exactly1 = num_have == 1
        have_atleast2 = num_have > 1

        assume_delta = 60 * 60 * 24  # 1 day
        if have_exactly1:
            # Only 1 date still means we have to extrapolate.
            have_idx = np.where(~missing_date)[0][0]
            have_time = unixtimes[have_idx]
            unixtimes = unixtimes.copy()
            unixtimes[0] = have_time - (have_idx * assume_delta)
            unixtimes[-1] = have_time + (len(unixtimes) - (have_idx + 1)) * assume_delta
            missing_date = np.isnan(unixtimes)
            have_atleast2 = True

        if have_atleast2:
            from scipy import interpolate
            frame_idxs = np.arange(len(unixtimes))
            miss_idxs = frame_idxs[missing_date]
            have_idxs = frame_idxs[~missing_date]
            have_values = unixtimes[have_idxs]
            interp = interpolate.interp1d(have_idxs, have_values, fill_value='extrapolate')
            interp_vals = interp(miss_idxs)
            unixtimes = unixtimes.copy()
            unixtimes[miss_idxs] = interp_vals
        else:
            # No information.
            unixtimes = np.linspace(0, len(unixtimes) * assume_delta, len(unixtimes))
    return unixtimes


def coerce_time_kernel(pattern):
    """
    Obtain a time kernel from user input

    Args:
        pattern (str | Iterable[str | Number]):
            A string code or a iterable of time coercable time deltas in
            ascending order. A pattern code is a ',' separated string of
            coercable time deltas.

    Returns:
        ndarray : ascending timedelta offsets in seconds

    Example:
        >>> from watch.tasks.fusion.datamodules.temporal_sampling.utils import *  # NOQA
        >>> import ubelt as ub
        >>> valid_patterns = [
        >>>     '-1y,-30d,-1d,0,1d,30d,1y',
        >>>     '-60s,0s,60s',
        >>>     '-1d,-60s,20s,60s,1d',
        >>>     '1,1,1,1,1',
        >>>     '(1,1,1,1,1)',
        >>> ]
        >>> for pattern in valid_patterns:
        >>>     kernel = coerce_time_kernel(pattern)
        >>>     assert np.all(kernel == coerce_time_kernel(pattern)), 'should be idempotent'
        >>>     print('kernel = {}'.format(ub.urepr(kernel.tolist(), nl=0)))
        kernel = [-31536000.0, -2592000.0, -86400.0, 0.0, 86400.0, 2592000.0, 31536000.0]
        kernel = [-60.0, 0.0, 60.0]
        kernel = [-86400.0, -60.0, 20.0, 60.0, 86400.0]
        kernel = [1.0, 1.0, 1.0, 1.0, 1.0]
        kernel = [1.0, 1.0, 1.0, 1.0, 1.0]
        >>> import pytest
        >>> invalid_patterns = [
        >>>     '3s,2s,1s'
        >>>     '-10,5,3,-2,0,1'
        >>> ]
        >>> for pattern in invalid_patterns:
        >>>     with pytest.raises(ValueError):
        >>>         kernel = coerce_time_kernel(pattern)
        >>> with pytest.raises(TypeError):
        >>>     coerce_time_kernel(3.14)
    """
    from watch.tasks.fusion.datamodules.temporal_sampling.time_kernel_grammar import parse_multi_time_kernel
    from watch.utils.util_time import coerce_timedelta
    if isinstance(pattern, str):
        if '(' in pattern:
            multi_kernel = parse_multi_time_kernel(pattern)
            assert len(multi_kernel) == 1, 'only expecting a single kernel here'
            kernel_deltas = multi_kernel[0]
        else:
            kernel_deltas = pattern.split(',')
    elif ub.iterable(pattern):
        kernel_deltas = pattern
    else:
        print(f'error: pattern={pattern}')
        raise TypeError(type(pattern))
    parsed = [coerce_timedelta(d) for d in kernel_deltas]
    kernel = np.array([v.total_seconds() for v in parsed])
    diffs = np.diff(kernel)
    if not np.all(diffs >= 0):
        raise ValueError('Inputs must be in ascending order')
    return kernel


def coerce_multi_time_kernel(pattern):
    """
    Obtain a list of time kernels from user input.

    Example:
        >>> from watch.tasks.fusion.datamodules.temporal_sampling.utils import *  # NOQA
        >>> import ubelt as ub
        >>> valid_patterns = [
        >>>     '(-1d,0,1d),(1,2)',
        >>>     '1,1,1',
        >>>     '(1,1,1,1,1)',
        >>>     ['-1,0,+1', '0'],
        >>> ]
        >>> for pattern in valid_patterns:
        >>>     multi_kernel = coerce_multi_time_kernel(pattern)
        >>>     recon = coerce_multi_time_kernel(multi_kernel)
        >>>     a = [r.tolist() for r in recon]
        >>>     b = [r.tolist() for r in multi_kernel]
        >>>     assert a == b, 'should be idempotent'
        >>>     print('multi_kernel = {}'.format(ub.urepr(multi_kernel, nl=1)))
        multi_kernel = [
            np.array([-86400.,      0.,  86400.], dtype=np.float64),
            np.array([1., 2.], dtype=np.float64),
        ]
        multi_kernel = [
            np.array([1., 1., 1.], dtype=np.float64),
        ]
        multi_kernel = [
            np.array([1., 1., 1., 1., 1.], dtype=np.float64),
        ]
        multi_kernel = [
            np.array([-1.,  0.,  1.], dtype=np.float64),
            np.array([0.], dtype=np.float64),
        ]
    """
    if pattern is None:
        return [None]
    from watch.tasks.fusion.datamodules.temporal_sampling.time_kernel_grammar import parse_multi_time_kernel
    from watch.utils.util_time import coerce_timedelta
    if isinstance(pattern, str):
        multi_kernel = parse_multi_time_kernel(pattern)
    elif ub.iterable(pattern):
        if len(pattern) == 0:
            multi_kernel = []
        else:
            first = pattern[0]
            if ub.iterable(first):
                # Assume we are given a list of pre-parsed kernels
                multi_kernel = pattern
                # multi_kernel = [coerce_timedelta(d) for d in pattern]
            else:
                # We might be given a list of multiple kernels
                try:
                    multi_kernel = [coerce_time_kernel(p) for p in pattern]
                except Exception:
                    # Or we might be given just a single kernel.
                    multi_kernel = [[coerce_timedelta(d) for d in pattern]]
                    ...

    else:
        print(f'error: pattern={pattern}')
        raise TypeError(type(pattern))
    return multi_kernel



import torch
import torch.nn as nn
from datetime import date
import random
from argparse import ArgumentParser, Namespace
import pytorch_lightning as pl
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor
import numpy as np


from .utils.attention_unet import attention_unet
from .data.datasets import kwcoco_dataset, SpaceNet7, Onera


class change(pl.LightningModule):
    def __init__(self, hparams):
        super().__init__()

        if type(hparams) == dict:
            hparams = Namespace(**hparams)

        ### Define backbone network.
        if hparams.pretrained_checkpoint:
            if hparams.pretrained_multihead:
                pretrained_model = self.load_from_checkpoint(hparams.pretrained_checkpoint)
            else:
                pretrained_model = self.load_from_checkpoint(hparams.pretrained_checkpoint)

            if hparams.pretrained_encoder_only:
                self.backbone.encoder = pretrained_model.backbone.encoder
            else:
                self.backbone = pretrained_model.backbone
        else:
            self.backbone = attention_unet(hparams.num_channels, hparams.feature_dim, pos_encode=hparams.positional_encoding, num_attention_layers=hparams.num_attention_layers, mode=hparams.positional_encoding_mode)

        ##### define dataset
        if hparams.trainset == 'kwcoco':
            self.trainset = kwcoco_dataset(hparams.train_kwcoco, hparams.sensor, hparams.bands, hparams.patch_size, segmentation_labels=True, num_images=hparams.num_images)
            self.valset = kwcoco_dataset(hparams.val_kwcoco, hparams.sensor, hparams.bands, hparams.patch_size, segmentation_labels=True, num_images=hparams.num_images)
        elif hparams.trainset == 'spacenet':
            self.trainset = SpaceNet7(hparams.patch_size, segmentation_labels=True, num_images=hparams.num_images, train=True)
            self.valset = SpaceNet7(hparams.patch_size, segmentation_labels=True, num_images=hparams.num_images, train=False)
        elif hparams.trainset == 'onera':
            self.trainset = Onera(train=True, patch_size=hparams.patch_size, num_channels=hparams.num_channels)
            self.valset = Onera(train=False, num_channels=hparams.num_channels)
            assert hparams.num_images == 2

        if hparams.binary:
            num_classes = 2
            weight = torch.FloatTensor([1, hparams.pos_class_weight])
        else:
            num_classes = 6
            weight = torch.FloatTensor([0, 1, 1, 1, 1, 1])

        self.criterion = nn.NLLLoss(weight=weight)
        self.save_hyperparameters(hparams)

        self.classifier = self.head(2 * hparams.feature_dim, num_classes)

    def head(self, in_channels, out_channels, kernel_size=3, dilation=1):
        return nn.Sequential(nn.Conv2d(in_channels, in_channels, kernel_size, dilation=dilation, bias=False, padding=dilation),
                             nn.GroupNorm(num_groups=in_channels // 8, num_channels=in_channels),
                             nn.ReLU(),
                             nn.Conv2d(in_channels, in_channels, kernel_size, dilation=dilation, bias=False, padding=dilation),
                             nn.GroupNorm(num_groups=in_channels // 8, num_channels=in_channels),
                             nn.ReLU(),
                             nn.Conv2d(in_channels, out_channels, 1, bias=False, padding=0),
                             nn.LogSoftmax(dim=1))

    def forward(self, x, positions=None):
        predictions = self.backbone(x, positions)
        #         _, predicted_class = torch.max(predictions, dim=2)
        return {
                'predictions': predictions,
                #                 'predicted_class': predicted_class
               }

    def shared_step(self, batch):
        images = [batch[key] for key in batch if key[:5] == 'image']

        if self.hparams.change:
            change_map = batch['change_map'].to(self.device)
        else:
            label = random.choice([0, 1])
            change_map = label * torch.ones_like(batch['change_map']).to(self.device)
            if not label:
                images.reverse()

        images = torch.stack(images, dim=1).to(self.device)

        if self.hparams.positional_encoding:
            positions = batch['time_steps'].to(self.device)
        else:
            positions = None

        forward = self.forward(images, positions)
        predictions = forward['predictions']

        predictions = self.classifier(torch.cat([predictions[:, 0, :, :, :], predictions[:, 1, :, :, :]], dim=1))

        loss = self.criterion(predictions, change_map.long())

        output = {
                    'prediction_map': predictions,
                    'targets' : change_map,
                    'loss': loss
                }

        return output

    def training_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        self.log('train_loss', output['loss'])
        return output['loss']

    def validation_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        self.log('validation_loss', output['loss'])
        return output['loss']

    def validation_epoch_end(self, outputs):
        if self.hparams.change:
            epoch_test_loss, epoch_test_accuracy, cl_acc, pr_rec = self.run_test(loader=self.val_dataloader)
            epoch_test_nochange_accuracy = cl_acc[0]
            epoch_test_change_accuracy = cl_acc[1]
            epoch_test_precision = pr_rec[0]
            epoch_test_recall = pr_rec[1]
            epoch_test_Fmeasure = pr_rec[2]
            self.log('val_epoch_accuracy', epoch_test_accuracy)
            self.log('val_epoch_accuracy_change', epoch_test_change_accuracy)
            self.log('val_epoch_accuracy_no_change', epoch_test_nochange_accuracy)
            self.log('val_epoch_precision', epoch_test_precision)
            self.log('val_epoch_recall', epoch_test_recall)
            self.log('val_epoch_f1', epoch_test_Fmeasure)

    def run_test(self, loader):
        self.eval()
        tot_loss = 0
        tot_count = 0

        n = 2
        class_correct = list(0. for i in range(n))
        class_total = list(0. for i in range(n))
        class_accuracy = list(0. for i in range(n))

        tp = 0
        tn = 0
        fp = 0
        fn = 0

        for batch in loader():
            images = [batch[key] for key in batch if key[:5] == 'image']
            images = torch.stack(images, dim=1).to(self.device)
            output = self.shared_step(batch)

            segmentations = output['targets']
            output = output['prediction_map']

            loss = self.criterion(output, segmentations.long())

            tot_loss += loss.data * np.prod(segmentations.size())
            tot_count += np.prod(segmentations.size())

            _, predicted = torch.max(output.data, 1)

            c = (predicted.int() == segmentations.data.int())

            where_no_change = (0 == segmentations.data.int())
            class_correct[0] += torch.sum(c[where_no_change])
            class_total[0] += torch.sum(where_no_change)
            where_change = (1 == segmentations.data.int())
            class_correct[1] += torch.sum(c[where_change])
            class_total[1] += torch.sum(where_change)

            pr = (predicted.int() > 0).cpu().numpy()
            gt = (segmentations.data.int() > 0).cpu().numpy()

            tp += np.logical_and(pr, gt).sum()
            tn += np.logical_and(np.logical_not(pr), np.logical_not(gt)).sum()
            fp += np.logical_and(pr, np.logical_not(gt)).sum()
            fn += np.logical_and(np.logical_not(pr), gt).sum()

        net_loss = tot_loss / tot_count
        net_accuracy = 100 * (tp + tn) / tot_count

        for i in range(n):
            class_accuracy[i] = 100 * class_correct[i] / max(class_total[i], 0.00001)
        prec = tp / (tp + fp)
        rec = tp / (tp + fn)
        f_meas = 2 * prec * rec / (prec + rec)
        prec_nc = tn / (tn + fn)
        rec_nc = tn / (tn + fp)

        pr_rec = [prec, rec, f_meas, prec_nc, rec_nc]

        return net_loss, net_accuracy, class_accuracy, pr_rec

    def train_dataloader(self):
        return torch.utils.data.DataLoader(self.trainset,
                                           batch_size=self.hparams.batch_size,
                                           num_workers=self.hparams.workers,
                                           shuffle=True)

    def val_dataloader(self):
        return torch.utils.data.DataLoader(self.valset,
                                           batch_size=1,
                                           num_workers=self.hparams.workers)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.learning_rate, weight_decay=self.hparams.weight_decay)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, self.hparams.step_size, gamma=self.hparams.lr_gamma)
#         lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, self.hparams.step_size)
        return {'optimizer': optimizer, 'lr_scheduler': lr_scheduler}


def main(args):
    if type(args) == dict:
        args = Namespace(**args)

    if args.change:
        task = 'change'
    else:
        task = 'before_after'

#     if args.pretrained_multihead and args.pretrained_checkpoint:
#         mh = 'pretrained_multihead'
#     elif args.pretrained_checkpoint:
#         mh = 'sort'
#     else:
#         mh = 'no_pretrain'

    if args.positional_encoding:
        mode = args.positional_encoding_mode
    else:
        mode = 'none'

    log_dir = '{}/{}/{}/{}/{}/{}/{}'.format(
        args.save_dir,
        args.trainset,
        task,
        'Attention_layers:' + str(args.num_attention_layers),
        'Position:' + str(args.positional_encoding),
        mode,
        str(date.today()),
        )

    logger = TensorBoardLogger(log_dir)

    model = change(hparams=args)

    checkpoint_callback = ModelCheckpoint(monitor='val_epoch_f1', mode='max', save_top_k=1)
    lr_logger = LearningRateMonitor(logging_interval='step')
    trainer = pl.Trainer.from_argparse_args(args,
                                            logger=logger,
                                            callbacks=[checkpoint_callback, lr_logger],
                                            log_every_n_steps=30,
                                            check_val_every_n_epoch=args.check_val_every_n_epoch)
    trainer.fit(model)


if __name__ == '__main__':

    parser = ArgumentParser()

    ###train hyperparameters
    parser.add_argument('--max_epochs', type=int, default=50)
    parser.add_argument('--workers', type=int, default=8)
    parser.add_argument('--batch_size', type=int, default=64)
    parser.add_argument('--step_size', type=int, default=10)
    parser.add_argument('--lr_gamma', type=float, default=.1)
    parser.add_argument('--weight_decay', type=float, default=1e-5)
    parser.add_argument('--learning_rate', type=float, default=.0001)
    parser.add_argument('--save_dir', default='watch/tasks/invariants/logs')
    parser.add_argument('--gpus', type=int, default=1)
    parser.add_argument('--feature_dim', type=int, default=64)
#     parser.add_argument('--drop_rate', type=float, default=.2)

    ###head network
    parser.add_argument('--kernel_size', type=int, default=3)
    parser.add_argument('--dilation', type=int, default=1)

    ###dataset
    parser.add_argument('--trainset', type=str, help='Choose from: spacenet, onera, or kwcoco.', default='onera')
    parser.add_argument('--valset', type=str, help='Choose from: spacenet, onera, or kwcoco. If blank, valset will correspond to chosen trainset.', default='onera')
    parser.add_argument('--patch_size', type=int, default=128)
    parser.add_argument('--num_channels', type=int, default=10)

    ### kwcoco arguments
    parser.add_argument('--train_kwcoco', type=str, default='')
    parser.add_argument('--val_kwcoco', type=str, default='')
    parser.add_argument('--sensor', type=str, nargs='+', default=['S2', 'L8'])
    parser.add_argument('--bands', type=str, nargs='+', default=['shared'])

    ### spacenet arguments
    parser.add_argument('--remove_clouds', help='spacenet specific argument', action='store_true')
    parser.add_argument('--normalize_spacenet', help='spacenet specific argument', action='store_true')
    parser.add_argument('--check_val_every_n_epoch', type=int, default=5)

    ### pretraining arguments
    parser.add_argument('--pretrained_checkpoint', type=str, help='path to pretrained checkpoint. Leave blank for change detection training without pretraining.', default='')
    parser.add_argument('--pretrained_multihead', action='store_true', help='indicate if the pretrained checkpoint was trained in a multihead fashion')
    parser.add_argument('--pretrained_encoder_only', action='store_true')

    ### main argument
    parser.add_argument('--binary', action='store_true')
    parser.add_argument('--pos_class_weight', type=float, help='Weight on positive class for segmentation. Only used on binary labels.', default=10)
    parser.add_argument('--num_images', type=int, default=2)
    parser.add_argument('--num_attention_layers', type=int, default=4)
    parser.add_argument('--positional_encoding', action='store_true')
    parser.add_argument('--positional_encoding_mode', type=str, help='addition or concatenation', default='concatenation')
    parser.add_argument('--change', action='store_true')

    args = parser.parse_args()
    main(args)



from watch.tasks.rutgers_material_seg_v2.models.material_mlp import MaterialMLP

MODELS = {'material_mlp': MaterialMLP}


def build_model(model_name, n_in_channels, n_out_channels, **kwargs):

    try:
        model_func = MODELS[model_name]
    except KeyError:
        raise NotImplementedError(f'Model "{model_name}" not implimented in build_model function.')

    model = model_func(n_in_channels, n_out_channels, **kwargs)

    return model



"""
Basline Example:

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)

    python -m watch.tasks.invariants.predict \
        --input_kwcoco=$DVC_DATA_DPATH/Drop4-BAS/data_vali_KR_R001.kwcoco.json \
        --output_kwcoco=$DVC_DATA_DPATH/Drop4-BAS/all_tests/model_thirteen_epoch/data_vali_KR_R001_invariants.kwcoco.json \
        --pretext_package=$DVC_EXPT_DPATH/models/uky/uky_invariants_2022_12_17/TA1_pretext_model/pretext_package.pt \
        --input_space_scale=10GSD  \
        --window_space_scale=10GSD \
        --patch_size=256 \
        --do_pca 0 \
        --patch_overlap=0.3 \
        --workers="2" \
        --write_workers 0 \
        --tasks before_after pretext

    # After your model predicts the outputs, you should be able to use the
    # smartwatch visualize tool to inspect your features.
    python -m watch visualize $DVC_DATA_DPATH/Drop4-BAS/all_tests/model_thirteen_epoch/data_vali_invariants.kwcoco.json \
        --channels "invariants.5:8,invariants.8:11,invariants.14:17" --stack=only --workers=avail --animate=True \
        --draw_anns=False


SeeAlso:
    ~/code/watch/watch/cli/prepare_teamfeats.py
"""
import torch
import ubelt as ub
from .data.datasets import GriddedDataset
from .pretext_model import pretext
from .segmentation_model import segmentation_model as seg_model
from watch.tasks.fusion.coco_stitcher import CocoStitchingManager
from watch.utils import util_kwimage  # NOQA
from watch.utils import util_parallel
from watch.utils.lightning_ext import util_device

import scriptconfig as scfg


class InvariantPredictConfig(scfg.DataConfig):
    """
    Configuration for UKY invariant models
    """
    device = scfg.Value('cuda', type=str)
    pretext_ckpt_path = scfg.Value(None, type=str)
    segmentation_ckpt_path = scfg.Value(None, type=str)
    pretext_package_path = scfg.Value(None, type=str)
    segmentation_package_path = scfg.Value(None, type=str)
    batch_size = scfg.Value(1, type=int)
    workers = scfg.Value(4, help=ub.paragraph(
            '''
            number of background data loading workers
            '''), alias=['num_workers'])
    io_workers = scfg.Value(0, help=ub.paragraph(
            '''
            number of background data writing workers
            '''), alias=['write_workers'])

    window_resolution = scfg.Value('10GSD', help='The window GSD to build the grid at', alias=['window_space_scale'])
    input_resolution = scfg.Value('10GSD', help='The input GSD to sample the grid at', alias=['input_space_scale'])

    sensor = scfg.Value(['S2', 'L8'], nargs='+')
    bands = scfg.Value(['shared'], type=str, help=ub.paragraph(
            '''
            Choose bands on which to train. Can specify 'all' for all
            bands from given sensor, or 'share' to use common bands when
            using both S2 and L8 sensors
            '''), nargs='+')
    patch_size = scfg.Value(256, type=int)
    patch_overlap = scfg.Value(0.25, type=float)
    input_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to kwcoco dataset with images to generate feature for
            '''))
    output_kwcoco = scfg.Value(None, type=str, required=True, help=ub.paragraph(
            '''
            Path to write an output kwcoco file. Output file will be a
            copy of input_kwcoco with addition feature fields generated
            by predict.py rerooted to point to the original data.
            '''))
    tasks = scfg.Value(['all'], help=ub.paragraph(
            '''
            Specify which tasks to choose from (segmentation,
            before_after, or pretext. Can also specify 'all')
            '''), nargs='+')
    do_pca = scfg.Value(1, type=int, help=ub.paragraph(
            '''
            Set to 1 to perform pca. Choose output dimension in num_dim
            argument.
            '''))
    pca_projection_path = scfg.Value('', type=str, help='Path to pca projection matrix')

    track_emissions = scfg.Value(True, help='Set to false to disable codecarbon')

    assets_dname = scfg.Value('_assets', help=ub.paragraph(
        '''
        The name of the top-level directory to write new assets.
        '''))

    def __post_init__(self):
        if 'all' in self.tasks:
            self['tasks'] = ['segmentation', 'before_after', 'pretext']


class Predictor(object):
    """
    CommandLine:
        DVC_DPATH=$(geowatch_dvc)
        DVC_DPATH=$DVC_DPATH xdoctest -m watch.tasks.invariants.predict Predictor

        python -m watch visualize $DVC_DPATH/Drop2-Aligned-TA1-2022-02-15/test_uky.kwcoco.json \
            --channels='invariants.0:3' --animate=True --with_anns=False

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from watch.tasks.invariants.predict import *  # NOQA
        >>> import kwcoco
        >>> import watch
        >>> dvc_dpath = watch.find_smart_dvc_dpath()
        >>> #  Write out smaller version of the dataset
        >>> dset = kwcoco.CocoDataset(dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/data_nowv_vali.kwcoco.json')
        >>> images = dset.videos(names=['KR_R001']).images[0]
        >>> sub_images = images.compress([s != 'WV' for s in images.lookup('sensor_coarse')])[::5]
        >>> sub_dset = dset.subset(sub_images)
        >>> sub_dset.fpath = (dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/small_test_data_nowv_vali.kwcoco.json')
        >>> sub_dset.dump(sub_dset.fpath)
        >>> input_kwcoco = sub_dset.fpath
        >>> output_kwcoco = dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/test_uky.kwcoco.json'
        >>> pretext_package_path = dvc_dpath / 'models/uky/uky_invariants_2022_03_11/TA1_pretext_model/pretext_package.pt'
        >>> pca_projection_path = dvc_dpath / 'models/uky/uky_invariants_2022_03_11/TA1_pretext_model/pca_projection_matrix.pt'
        >>> segmentation_package_path = dvc_dpath / 'models/uky/uky_invariants_2022_02_11/TA1_segmentation_model/segmentation_package.pt'
        >>> argv = []
        >>> argv += ['--input_kwcoco', f'{sub_dset.fpath}']
        >>> argv += ['--output_kwcoco', f'{output_kwcoco}']
        >>> argv += ['--pca_projection_path', f'{pca_projection_path}']
        >>> argv += ['--pretext_package_path', f'{pretext_package_path}']
        >>> argv += ['--segmentation_package_path', f'{segmentation_package_path}']
        >>> argv += ['--patch_overlap', '0.25']
        >>> argv += ['--workers', '2']
        >>> argv += ['--tasks', 'all']
        >>> argv += ['--do_pca', '1']
        >>> args = InvariantPredictConfig.cli(argv=argv)
        >>> self = Predictor(args)
        >>> self.forward(args)
    """

    def __init__(self, args):

        # TODO: Add a cache flag. If cache==1, Determine what images we have
        # already predicted for and then take a kwcoco subset containing
        # the cache misses. See dzyne and rutgers predictors for example
        # implementations.

        # Doesnt work?
        # FIX_ALBUMENTATIONS_HACK = 0
        # if FIX_ALBUMENTATIONS_HACK:
        #     from albumentations.core import composition
        #     composition.Transforms = composition.TransformType

        self.batch_size = args.batch_size
        self.do_pca = args.do_pca

        self.tasks = args.tasks

        self.workers = util_parallel.coerce_num_workers(args.workers)
        print('self.workers = {!r}'.format(self.workers))

        self.io_workers = util_parallel.coerce_num_workers(args.io_workers)
        print(f'self.io_workers={self.io_workers}')

        self.devices = util_device.coerce_devices(args.device)
        assert len(self.devices) == 1, 'only 1 for now'
        self.device = device = self.devices[0]
        print('device = {!r}'.format(device))

        # Initialize models
        print('Initialize models')
        if 'all' in args.tasks:
            self.tasks = ['segmentation', 'before_after', 'pretext']
        else:
            self.tasks = args.tasks
        ### Define tasks
        if 'segmentation' in self.tasks:
            if args.segmentation_package_path:
                print('Initialize segmentation model from package')
                self.segmentation_model = seg_model.load_package(args.segmentation_package_path)
            else:
                print('Initialize segmentation model from checkpoint')
                self.segmentation_model = seg_model.load_from_checkpoint(args.segmentation_ckpt_path, dataset=None)
            self.segmentation_model = self.segmentation_model.to(device)

        if 'pretext' in self.tasks:
            if args.pretext_package_path:
                print('Initialize pretext model from package')
                self.pretext_model = pretext.load_package(args.pretext_package_path)
            else:
                print('Initialize pretext model from checkpoint')
                self.pretext_model = pretext.load_from_checkpoint(args.pretext_ckpt_path, train_dataset=None, vali_dataset=None)
            self.pretext_model = self.pretext_model.eval().to(device)
            # pretext_hparams = pretext_model.hparams

            try:
                # Hack
                self.pretext_model.sort_accuracy = None
            except Exception:
                pass
            self.pretext_model.__dict__['sort_accuracy'] = ub.identity  # HUGE HACK

        self.in_feature_dims = self.pretext_model.hparams.feature_dim_shared
        if args.do_pca:
            print('Initialize PCA model')
            self.pca_projector = torch.load(args.pca_projection_path).to(device)
            self.out_feature_dims = self.pca_projector.shape[0]
        else:
            self.out_feature_dims = self.in_feature_dims

        self.num_out_channels = self.out_feature_dims
        if 'segmentation' in self.tasks:
            self.num_out_channels += 1
        if 'before_after' in self.tasks:
            self.num_out_channels += 1

        # initialize dataset
        import kwcoco
        print('load coco dataset')
        self.coco_dset = kwcoco.CocoDataset = kwcoco.CocoDataset.coerce(args.input_kwcoco)
        # self.coco_dset = self.coco_dset.subset(list(self.coco_dset.images()[0:20]))

        ###
        print('build grid dataset')
        self.dataset = GriddedDataset(self.coco_dset,
                                      sensor=args.sensor,
                                      bands=args.bands,
                                      patch_size=args.patch_size,
                                      patch_overlap=args.patch_overlap,
                                      window_space_scale=args.window_resolution,
                                      input_space_scale=args.input_resolution,
                                      output_space_scale=args.input_resolution,
                                      mode='test')

        print('copy dataset')
        self.output_dset = self.dataset.coco_dset.copy()

        print('reroot')
        self.output_dset.reroot(absolute=True)  # Make all paths absolute
        self.output_dset.fpath = args.output_kwcoco  # Change output file path and bundle path
        self.output_dset.reroot(absolute=False)  # Reroot in the new bundle path
        self.finalized_gids = set()
        # self.stitcher_dict = {}

        self.save_channels = f'invariants:{self.num_out_channels}'
        self.output_kwcoco_path = ub.Path(args.output_kwcoco)

        self.imwrite_kw = {
            'compress': 'DEFLATE',
            'backend': 'gdal',
            'blocksize': 128,
        }

        self.stitch_manager = CocoStitchingManager(
            result_dataset=self.output_dset,
            short_code='pred_invariants',
            chan_code=self.save_channels,
            stiching_space='video',
            prob_compress=self.imwrite_kw['compress'],
            quantize=True,
            assets_dname=args.assets_dname,
        )

        from watch.utils import process_context
        self.proc_context = process_context.ProcessContext(
            type='process',
            name='watch.tasks.invariants.predict',
            config=args.to_dict(),
            track_emissions=args.track_emissions,
        )

    def forward(self):
        device = self.device

        loader = torch.utils.data.DataLoader(
            self.dataset, num_workers=self.workers,
            batch_size=self.batch_size, shuffle=False)
        num_batches = len(loader)

        # Start background processes
        # Build a task queue for background write results workers
        from watch.utils import util_parallel
        writer_queue = util_parallel.BlockingJobQueue(max_workers=self.io_workers)
        self.stitch_manager.writer_queue = writer_queue

        self.proc_context.start()

        self.proc_context.add_disk_info(ub.Path(self.output_dset.fpath).parent)
        self.output_dset.dataset.setdefault('info', [])
        self.output_dset.dataset['info'].append(self.proc_context.obj)

        print('Evaluating and saving features')

        with torch.set_grad_enabled(False):
            seen_images = set()
            prog = ub.ProgIter(enumerate(loader), total=num_batches, desc='Compute features', verbose=1)
            for idx, batch in prog:
                save_feat = []
                save_feat2 = []

                # Handle input nans
                img1 = batch['image1']
                img2 = batch['image2']
                offset_image1 = batch['offset_image1']
                augmented_image1 = batch['augmented_image1']

                invalid_mask1 = torch.isnan(img1)[0].any(dim=0)
                invalid_mask2 = torch.isnan(img2)[0].any(dim=0)

                batch['image1'] = torch.nan_to_num(img1).to(device)
                batch['image2'] = torch.nan_to_num(img2).to(device)
                batch['offset_image1'] = torch.nan_to_num(offset_image1).to(device)
                batch['augmented_image1'] = torch.nan_to_num(augmented_image1).to(device)

                if 'pretext' in self.tasks:

                    image_stack = torch.stack([batch['image1'], batch['image2'], batch['offset_image1'], batch['augmented_image1']], dim=1)
                    image_stack = image_stack.to(device)

                    # Remove nans before going into the network
                    image_stack = torch.nan_to_num(image_stack)

                    #select features corresponding to first image
                    features = self.pretext_model(image_stack)[:, 0, :, :, :]
                    #select features corresponding to second image
                    features2 = self.pretext_model(image_stack)[:, 1, :, :, :]
                    if self.do_pca:
                        features = torch.einsum('xy,byhw->bxhw', self.pca_projector, features)
                        features2 = torch.einsum('xy,byhw->bxhw', self.pca_projector, features2)

                    features = features.squeeze().permute(1, 2, 0).cpu()
                    features2 = features2.squeeze().permute(1, 2, 0).cpu()

                    features[invalid_mask1] = float('nan')
                    features2[invalid_mask2] = float('nan')

                    save_feat.append(features)
                    save_feat2.append(features2)

                if 'before_after' in self.tasks:
                    ### TO DO: Set to output of separate model.
                    before_after_heatmap = self.pretext_model.shared_step(batch)['before_after_heatmap'][0].permute(1, 2, 0)
                    before_after_heatmap = torch.sigmoid(torch.exp(before_after_heatmap[:, :, 1]) - torch.exp(before_after_heatmap[:, :, 0])).unsqueeze(-1).cpu()

                    before_after_heatmap[invalid_mask1] = float('nan')
                    before_after_heatmap[invalid_mask2] = float('nan')

                    save_feat.append(before_after_heatmap)
                    save_feat2.append(before_after_heatmap)

                if 'segmentation' in self.tasks:
                    image_stack = [batch[key] for key in batch if key.startswith('image')]
                    image_stack = torch.stack(image_stack, dim=1).to(self.device)
                    predictions = torch.exp(self.segmentation_model(image_stack)['predictions'])

                    segmentation_heatmap = torch.sigmoid(predictions[0, 0, 1, :, :] - predictions[0, 0, 0, :, :]).unsqueeze(0).permute(1, 2, 0).cpu()
                    segmentation_heatmap2 = torch.sigmoid(predictions[0, 1, 1, :, :] - predictions[0, 1, 0, :, :]).unsqueeze(0).permute(1, 2, 0).cpu()

                    segmentation_heatmap[invalid_mask1] = float('nan')
                    segmentation_heatmap2[invalid_mask2] = float('nan')

                    save_feat.append(segmentation_heatmap)
                    save_feat2.append(segmentation_heatmap2)

                save_feat = torch.cat(save_feat, dim=-1)
                save_feat = save_feat.numpy()
                save_feat2 = torch.cat(save_feat2, dim=-1)
                save_feat2 = save_feat2.numpy()

                target = self.dataset.patches[idx]

                # These dataloader has told us that these iamges are now
                # complete, and thus can be finalized free any memory used by
                # its stitcher
                new_complete_gids = target.get('new_complete_gids', [])
                for gid in new_complete_gids:
                    assert gid not in seen_images
                    seen_images.add(gid)
                    # if 1:
                    #     img = self.dataset.coco_dset.index.imgs[gid]
                    #     frame_index = img['frame_index']
                    #     video_id = img['video_id']
                    #     prog.ensure_newline()
                    #     print(f'finalize {video_id=}, {gid=}, {frame_index=}')
                    # writer_queue.submit(self.stitch_manager.finalize, gid)
                    self.stitch_manager.submit_finalize_image(gid)

                gid1, gid2 = target['gids']
                # slice_ = target['space_slice']
                # stitcher1 = self.ensure_stitcher(gid1)
                # stitcher2 = self.ensure_stitcher(gid2)
                # CocoStitchingManager._stitcher_center_weighted_add(
                #     stitcher1, slice_, save_feat)
                # CocoStitchingManager._stitcher_center_weighted_add(
                #     stitcher2, slice_, save_feat2)

                sample_outspace_ltrb = util_kwimage.Box.coerce(batch['sample_outspace_ltrb'].numpy(), format='ltrb')
                full_stitch_outspace_box = util_kwimage.Box.coerce(batch['full_stitch_outspace_ltrb'].numpy(), format='ltrb')
                scale_outspace_from_vid = batch['scale_outspace_from_vid'].numpy()[0]
                outspace_slice = sample_outspace_ltrb.to_slice()
                outspace_dsize = full_stitch_outspace_box.dsize

                self.stitch_manager.accumulate_image(
                    gid1, outspace_slice, save_feat,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

                self.stitch_manager.accumulate_image(
                    gid2, outspace_slice, save_feat2,
                    dsize=outspace_dsize,
                    scale=scale_outspace_from_vid)

            print('Finalize already compelted jobs')
            writer_queue.wait_until_finished(desc='Finalize submitted jobs')

            # Finalize everything else that hasn't completed
            for gid in ub.ProgIter(list(self.stitch_manager.image_stitchers.keys()), desc='submit loose write jobs'):
                if gid not in seen_images:
                    seen_images.add(gid)
                    self.stitch_manager.submit_finalize_image(gid)
                    # writer_queue.submit(self.stitch_manager.finalize, gid)

            print('Finalize loose jobs')
            writer_queue.wait_until_finished()

        print('Finish process context')
        self.proc_context.add_device_info(device)
        self.proc_context.stop()

        print('Write to dset.fpath = {!r}'.format(self.output_dset.fpath))
        self.output_dset.dump(self.output_dset.fpath, newlines=True)
        print('Done')


def main():
    args = InvariantPredictConfig.cli()
    Predictor(args).forward()


if __name__ == '__main__':
    """
    SeeAlso:
        ../../cli/prepare_teamfeats.py

        # Team Features on Drop3
        DVC_DPATH=$(geowatch_dvc)
        KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Aligned-Drop3-TA1-2022-03-10
        python -m watch.cli.prepare_teamfeats \
            --base_fpath=$KWCOCO_BUNDLE_DPATH/data.kwcoco.json \
            --with_depth=0 \
            --with_landcover=0 \
            --with_materials=0  \
            --with_invariants=1 \
            --do_splits=0 \
            --gres=0 --backend=serial --run=1

    CommandLine:
        python -m watch.tasks.template.predict --help

        DVC_DPATH=$(geowatch_dvc)
        PRETEXT_PATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_pretext_model/pretext_package.pt
        SSEG_PATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_segmentation_model/segmentation_package.pt
        PCA_FPATH=$DVC_DPATH/models/uky/uky_invariants_2022_02_11/TA1_pretext_model/pca_projection_matrix.pt
        KWCOCO_BUNDLE_DPATH=$DVC_DPATH/Drop2-Aligned-TA1-2022-02-15
        python -m watch.tasks.invariants.predict \
            --pretext_package_path "$PRETEXT_PATH" \
            --segmentation_package_path "$SSEG_PATH" \
            --pca_projection_path "$PCA_FPATH" \
            --input_kwcoco $KWCOCO_BUNDLE_DPATH/data.kwcoco.json \
            --workers=avail \
            --do_pca 0 \
            --patch_overlap=0.3 \
            --output_kwcoco $KWCOCO_BUNDLE_DPATH/uky_invariants.kwcoco.json \
            --tasks before_after pretext

        python -m watch stats $KWCOCO_BUNDLE_DPATH/uky_invariants.kwcoco.json

        python -m watch visualize $KWCOCO_BUNDLE_DPATH/uky_invariants/invariants_nowv_vali.kwcoco.json \
            --channels "invariants.7,invariants.6,invariants.5" --animate=True \
            --select_images '.sensor_coarse != "WV"' --draw_anns=False

    Ignore:
        ### Command 1 / 2 - watch-teamfeat-job-0
        python -m watch.tasks.invariants.predict \
            --input_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/data_kr1br2.kwcoco.json" \
            --output_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/data_kr1br2_uky_invariants.kwcoco.json" \
            --pretext_package_path "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_package.pt" \
            --pca_projection_path  "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_pca_104.pt" \
            --do_pca 0 \
            --patch_overlap=0.0 \
            --workers="2" \
            --io_workers 0 \
            --tasks before_after pretext

        cd /home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC
        kwcoco subset --src=data.kwcoco.json --dst=AE_R001.kwcoco.json --select_videos='.name == "AE_R001"'
        kwcoco subset --src=data.kwcoco.json --dst=NZ_R001.kwcoco.json --select_videos='.name == "NZ_R001"'

        python -m watch.tasks.invariants.predict \
            --input_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/KR_R001.kwcoco.json" \
            --output_kwcoco "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC/KR_R001_invariants.kwcoco.json" \
            --pretext_package_path "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_package.pt" \
            --pca_projection_path  "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/uky/uky_invariants_2022_03_21/pretext_model/pretext_pca_104.pt" \
            --input_space_scale=30GSD \
            --window_space_scale=30GSD \
            --patch_size=256 \
            --do_pca 0 \
            --patch_overlap=0.0 \
            --workers="2" \
            --io_workers 2 \
            --tasks before_after pretext

        python -m watch visualize KR_R001_invariants.kwcoco.json \
            --channels "invariants.5:8,invariants.8:11,invariants.14:17" --stack=only --workers=avail --animate=True \
            --draw_anns=False


    """
    main()



"""
This is step 3/4 in predict.py


SeeAlso:

    predict.py

    prepare_kwcoco.py

    tile_processing_kwcoco.py

    export_cold_result_kwcoco.py *

    assemble_cold_result_kwcoco.py

This script is for exporting COLD algorithm results (change vector, coefficients, RMSEs)
to geotiff raster with kwcoco dataset.
See original code: ~/code/pycold/src/python/pycold/imagetool/export_change_map.py
"""

import os
import numpy as np
import pandas as pd
import datetime as datetime_mod
import json
import scriptconfig as scfg
import logging
import itertools
import ubelt as ub
import pytz
from datetime import datetime as datetime_cls
import gc
import kwcoco
logger = logging.getLogger(__name__)


try:
    from xdev import profile
except ImportError:
    from ubelt import identity as profile


class ExportColdKwcocoConfig(scfg.DataConfig):
    """
    The docstring will be the description in the CLI help
    """
    rank = scfg.Value(None, help='rank id')
    n_cores = scfg.Value(None, help='total cores assigned')
    stack_path = scfg.Value(None, help='folder directory of stacked data')
    reccg_path = scfg.Value(
        None, help='folder directory of cold processing result')
    meta_fpath = scfg.Value(
        None, help='file path of metadata json created by prepare_kwcoco script')
    combined_coco_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        a path to a file to combined kwcoco file
        '''))
    year_lowbound = scfg.Value(
        None, help='min year for saving geotiff, e.g., 2017')
    year_highbound = scfg.Value(
        None, help='max year for saving geotiff, e.g., 2022')
    coefs = scfg.Value(
        None,
        type=str,
        help="list of COLD coefficients for saving geotiff, e.g., a0,c1,a1,b1,a2,b2,a3,b3,cv,rmse")
    coefs_bands = scfg.Value(
        None,
        type=str,
        help='indicate the ba_nds for output coefs_bands, e.g., 0,1,2,3,4,5')
    timestamp = scfg.Value(
        False,
        help='True: exporting cold result by timestamp, False: exporting cold result by year, Default is False')
    combine = scfg.Value(True, help='for temporal combined mode, Default is True')
    sensors = scfg.Value('L8', type=str, help='sensor type, default is "L8"')


@profile
def export_cold_main(cmdline=1, **kwargs):
    """_summary_

    Args:
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m watch.tasks.cold.export_cold_result_kwcoco --help
        TEST_COLD=1 xdoctest -m watch.tasks.cold.export_cold_result_kwcoco export_cold_main

    Example:
    >>> # xdoctest: +REQUIRES(env:TEST_COLD)
    >>> from watch.tasks.cold.export_cold_result_kwcoco import export_cold_main
    >>> from watch.tasks.cold.export_cold_result_kwcoco import *
    >>> kwargs= dict(
    >>>    rank = 0,
    >>>    n_cores = 1,
    >>>    stack_path = "/home/jws18003/data/dvc-repos/smart_data_dvc/Drop6/_pycold_combine/stacked/KR_R001/",
    >>>    reccg_path = "/home/jws18003/data/dvc-repos/smart_data_dvc/Drop6/_pycold_combine/reccg/KR_R001/",
    >>>    meta_fpath = '/home/jws18003/data/dvc-repos/smart_data_dvc/Drop6/_pycold_combine/stacked/KR_R001/block_x9_y9/crop_20210807T010000Z_N37.643680E128.649453_N37.683356E128.734073_L8_0.json',
    >>>    combined_coco_fpath = "/home/jws18003/data/dvc-repos/smart_data_dvc/Drop6_MeanYear/data_vali_split1_KR_R001_MeanYear.kwcoco.json",
    >>>    coefs = 'cv,rmse,a0,a1,b1,c1',
    >>>    year_lowbound = None,
    >>>    year_highbound = None,
    >>>    coefs_bands = '0,1,2,3,4,5',
    >>>    timestamp = False,
    >>>    combine = True,
    >>>    )
    >>> cmdline=0
    >>> export_cold_main(cmdline, **kwargs)
    """
    pman = kwargs.pop('pman', None)
    config_in = ExportColdKwcocoConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    rank = config_in['rank']
    n_cores = config_in['n_cores']
    stack_path = ub.Path(config_in['stack_path'])
    reccg_path = ub.Path(config_in['reccg_path'])
    meta_fpath = ub.Path(config_in['meta_fpath'])
    year_lowbound = config_in['year_lowbound']
    year_highbound = config_in['year_highbound']
    coefs = config_in['coefs']
    coefs_bands = config_in['coefs_bands']
    combine = config_in['combine']
    timestamp = config_in['timestamp']
    sensors = config_in['sensors']

    if combine:
        combined_coco_fpath = ub.Path(config_in['combined_coco_fpath'])
    else:
        combined_coco_fpath = None

    # assert (combine and not timestamp) or (not combine and timestamp), "combine and timestamp must be either True and False, or False and True."
    # TODO: MPI mode
    # if config_in['rank'] == 'MPI':
    #     ## MPI mode
    #     raise NotImplementedError('todo')
    #     MPI = 'TODO'
    #     comm = MPI.COMM_WORLD
    #     rank = comm.Get_rank()
    #     n_cores = comm.Get_size()
    # else:
    #     rank = config_in['rank']
    #     n_cores = config_in['n_cores']

    # define variables
    config = json.loads(meta_fpath.read_text())
    n_cols = config['padded_n_cols']
    n_rows = config['padded_n_rows']
    n_block_x = config['n_block_x']
    n_block_y = config['n_block_y']
    block_width = int(n_cols / n_block_x)  # width of a block
    block_height = int(n_rows / n_block_y)  # height of a block
    n_blocks = n_block_x * n_block_y  # total number of blocks

    cold_param = json.loads((reccg_path / 'log.json').read_text())
    method = cold_param['algorithm']

    # coef_names = ['cv', 'rmse', 'a0', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'c1']
    # band_names = [0, 1, 2, 3, 4, 5]

    if coefs is not None:
        try:
            coefs = list(coefs.split(","))
        except Exception:
            print(f'coefs={coefs}')
            print(
                "Illegal coefs inputs: example, --coefs='a0, c1, a1, b1, a2, b2, a3, b3, cv, rmse'")
            raise

        try:
            coefs_bands = list(coefs_bands.split(","))
            coefs_bands = [int(coefs_band) for coefs_band in coefs_bands]
        except Exception:
            print(f'coefs_bands={coefs_bands}')
            print(
                "Illegal coefs_bands inputs: example, --coefs_bands='0, 1, 2, 3, 4, 5, 6'")
            raise

    dt = np.dtype([('t_start', np.int32),
                   ('t_end', np.int32),
                   ('t_break', np.int32),
                   ('pos', np.int32),
                   ('num_obs', np.int32),
                   ('category', np.short),
                   ('change_prob', np.short),
                   # note that the slope coefficient was scaled up by 10000
                   ('coefs', np.float32, (7, 8)),
                   ('rmse', np.float32, 7),
                   ('magnitude', np.float32, 7)])

    # if coefs is not None:
    #     assert all(elem in coef_names for elem in coefs)
    #     assert all(elem in band_names for elem in coefs_bands)

    out_path = reccg_path / 'cold_feature'
    tmp_path = reccg_path / 'cold_feature' / 'tmp'

    tz = pytz.timezone('US/Eastern')

    if rank == 0:
        out_path.ensuredir()
        tmp_path.ensuredir()

    # MPI mode
    # trans = comm.bcast(trans, root=0)
    # proj = comm.bcast(proj, root=0)
    # cols = comm.bcast(cols, root=0)
    # rows = comm.bcast(rows, root=0)
    # config = comm.bcast(config, root=0)

    # Get ordinal list from sample block_folder
    block_folder = stack_path / 'block_x1_y1'
    meta_files = [m for m in os.listdir(block_folder) if m.endswith('.json')]

    # sort image files by ordinal dates
    img_dates = []
    img_names = []

    # read metadata and
    for meta in meta_files:
        meta_config = json.loads((block_folder / meta).read_text())
        ordinal_date = meta_config['ordinal_date']
        img_name = meta_config['image_name'] + '.npy'
        img_dates.append(ordinal_date)
        img_names.append(img_name)

    if year_lowbound is None:
        year_low_ordinal = min(img_dates)
        year_lowbound = pd.Timestamp.fromordinal(year_low_ordinal).year
    else:
        year_low_ordinal = pd.Timestamp.toordinal(
            datetime_mod.datetime(int(year_lowbound), 1, 1))

    if year_highbound is None:
        year_high_ordinal = max(img_dates)
        year_highbound = pd.Timestamp.fromordinal(year_high_ordinal).year
    else:
        year_high_ordinal = pd.Timestamp.toordinal(
            datetime_mod.datetime(int(year_highbound + 1), 1, 1))

    img_dates, img_names = zip(*filter(lambda x: x[0] >= year_low_ordinal,
                                       zip(img_dates, img_names)))
    img_dates, img_names = zip(*filter(lambda x: x[0] < year_high_ordinal,
                                       zip(img_dates, img_names)))

    img_dates = sorted(img_dates)
    img_names = sorted(img_names)
    if timestamp:
        ordinal_day_list = img_dates
    if combine:
        combined_coco_dset = kwcoco.CocoDataset(combined_coco_fpath)

        # filter by sensors
        all_images = combined_coco_dset.images(list(ub.flatten(combined_coco_dset.videos().images)))
        flags = [s in sensors for s in all_images.lookup('sensor_coarse')]
        all_images = all_images.compress(flags)
        image_id_iter = iter(all_images)

        # Get ordinal date of combined coco image
        ordinal_dates = []
        ordinal_dates_july_first = []
        for image_id in image_id_iter:
            combined_coco_image: kwcoco.CocoImage = combined_coco_dset.coco_image(image_id)
            ts = combined_coco_image.img['timestamp']
            timestamp_local = datetime_cls.fromtimestamp(ts, tz=tz)
            timestamp_utc = timestamp_local.astimezone(pytz.utc)
            july_first = datetime_mod.date(timestamp_utc.year, 7, 1).toordinal()
            ordinal = timestamp_utc.toordinal()
            ordinal_dates.append(ordinal)
            ordinal_dates_july_first.append(july_first)

        ordinal_day_list = ordinal_dates

        # Back up
    if not timestamp:
        first_ordinal_dates = []
        first_img_names = []
        last_year = None
        for ordinal_day, img_name in zip(img_dates, img_names):
            year = pd.Timestamp.fromordinal(ordinal_day).year
            if year != last_year:
                first_ordinal_dates.append(ordinal_day)
                first_img_names.append(img_name)
                last_year = year
        ordinal_day_list = first_ordinal_dates

    ranks_percore = int(np.ceil(n_blocks / n_cores))
    i_iter = range(ranks_percore)
    if pman is not None:
        i_iter = pman.progiter(i_iter, total=ranks_percore,
                               desc=f'Export COLD \\[rank {rank}]', transient=True)

    for i in i_iter:
        iblock = n_cores * i + rank
        if iblock >= n_blocks:
            break
        current_block_y = int(np.floor(iblock / n_block_x)) + 1
        current_block_x = iblock % n_block_x + 1
        if method == 'OBCOLD':
            filename = f'record_change_x{current_block_x}_y{current_block_y}_obcold.npy'
        elif method == 'COLD':
            filename = f'record_change_x{current_block_x}_y{current_block_y}_cold.npy'
        elif method == 'HybridCOLD':
            filename = f'record_change_x{current_block_x}_y{current_block_y}_hybridcold.npy'

        block_folder = stack_path / \
            f'block_x{current_block_x}_y{current_block_y}'
        reccg_fpath = reccg_path / filename

        now = datetime_cls.now(tz).strftime('%Y-%m-%d %H:%M:%S')
        print(f'processing the rec_cg file {reccg_fpath} ({now})')

        # if not reccg_fpath.exists():
        #     print(f'the rec_cg file {reccg_fpath} is missing')

        if coefs is not None:
            results_block_coefs = np.full(
                (block_height, block_width, len(coefs) * len(coefs_bands),
                    len(ordinal_day_list)), -9999, dtype=np.float32)

            status_fpath = reccg_fpath.with_suffix('.status.json')
            if status_fpath.exists():
                with open(status_fpath, 'r') as f:
                    status_data = json.load(f)
                if status_data.get('status') == 'failed':
                    print(f'the rec_cg file {reccg_fpath} has failed status')
                    results_block_coefs
                else:
                    cold_block = np.array(np.load(reccg_fpath), dtype=dt)
                    cold_block_split = np.split(cold_block, np.argwhere(np.diff(cold_block['pos']) != 0)[:, 0] + 1)

                    nan_val = -9999
                    feature_outputs = coefs
                    feature_set = set(feature_outputs)

                    if not feature_set.issubset({'a0', 'c1', 'a1', 'b1', 'a2', 'b2', 'a3', 'b3', 'cv', 'rmse'}):
                        raise Exception('the outputted feature must be in [a0, c1, a1, b1, a2, b2, a3, b3, cv, rmse]')

                    for element in cold_block_split:
                        # Degugging mode
                        # if element[0]["pos"] == 4:
                        #     print(element)
                        # the relative column number in the block
                        i_col = int((element[0]["pos"] - 1) % n_cols) - \
                                (current_block_x - 1) * block_width
                        i_row = int((element[0]["pos"] - 1) / n_cols) - \
                                (current_block_y - 1) * block_height

                        for band_idx, band in enumerate(coefs_bands):
                            feature_row = extract_features(element, band, ordinal_day_list, nan_val, timestamp,
                                                            feature_outputs, feature_set)
                            # Degugging mode
                            # if current_block_x == 1 and current_block_y == 1:
                            #     if i_col == 3 and i_row == 0:
                            #         print(element[0]["pos"])
                            #         print((current_block_x, current_block_y), (i_col, i_row), (band_idx, band))
                            #         print('feature_row', feature_row)

                            for index, coef in enumerate(coefs):
                                # Degugging mode
                                # if i_col == 3 and i_row == 0 and element[0]["pos"] == 4:
                                #     print(feature_row[index])
                                results_block_coefs[i_row][i_col][index + band_idx * len(coefs)][:] = \
                                    feature_row[index]

        # save the temp dataset out
        for day in range(len(ordinal_day_list)):
            if coefs is not None:
                outfile = tmp_path / \
                    f'tmp_coefmap_block{iblock + 1}_{ordinal_day_list[day]}.npy'
                np.save(outfile, results_block_coefs[:, :, :, day])
    # MPI mode (wait for all processes)
    # comm.Barrier()

    gc.collect()


class NoMatchingColdCurve(Exception):
    ...


@profile
def extract_features(cold_plot, band, ordinal_day_list,
                     nan_val, timestamp, feature_outputs, feature_set):
    # NOTE: this function is a bottleneck, speedups are needed here

    features = np.full(
        (len(feature_outputs),
         len(ordinal_day_list)),
        nan_val,
        dtype=np.double)
    SLOPE_SCALE = 10000

    max_days_list = []
    for i in range(len(cold_plot)):
        last_year = pd.Timestamp.fromordinal(cold_plot[i]['t_end']).year
        max_days_list.append(datetime_mod.date(last_year, 12, 31).toordinal())

    # last_year = pd.Timestamp.fromordinal(cold_plot[i]['t_end']).year
    # max_days_list = [datetime_mod.date(last_year, 12, 31).toordinal()] * len(cold_plot)

    break_year_list = [-9999 if not (curve['t_break'] > 0 and curve['change_prob'] == 100) else
                       pd.Timestamp.fromordinal(curve['t_break']).year for curve in cold_plot]

    possible_features = ['a0', 'c1', 'a1', 'b1', 'rmse', 'cv']
    fk_to_idx = {
        fk: feature_outputs.index(fk)
        for fk in possible_features
        if fk in feature_set
    }

    a0_idx = fk_to_idx.get('a0', None)
    c1_idx = fk_to_idx.get('c1', None)
    a1_idx = fk_to_idx.get('a1', None)
    b1_idx = fk_to_idx.get('b1', None)
    rmse_idx = fk_to_idx.get('rmse', None)
    cv_idx = fk_to_idx.get('cv', None)

    # Precompute as much as possible before running the product
    idx_day_year_list = [
        (day_idx, ordinal_day, pd.Timestamp.fromordinal(ordinal_day).year)
        for day_idx, ordinal_day in enumerate(ordinal_day_list)
    ]

    mday_byear_curve_list = [
        (max_days_list[idx], break_year_list[idx], cold_curve)
        for idx, cold_curve in enumerate(cold_plot)]

    try:
        idxs_iter = itertools.product(idx_day_year_list, mday_byear_curve_list)
        for (day_idx, ordinal_day, ord_year), (max_day,
                                               break_year, cold_curve) in idxs_iter:
            if cv_idx is not None:
                if cold_curve['t_break'] != 0 and cold_curve['change_prob'] == 100:
                    if (timestamp and ordinal_day == cold_curve['t_break']) or (
                            not timestamp and break_year == ord_year):
                        features[cv_idx, day_idx] = cold_curve['magnitude'][band]
            if cold_curve['t_start'] <= ordinal_day < max_day:
                if a0_idx is not None:
                    features[a0_idx, day_idx] = (
                        cold_curve['coefs'][band][0] +
                        cold_curve['coefs'][band][1] *
                        ordinal_day / SLOPE_SCALE)
                if c1_idx is not None:
                    features[c1_idx, day_idx] = cold_curve['coefs'][band][1] / SLOPE_SCALE
                if a1_idx is not None:
                    features[a1_idx, day_idx] = cold_curve['coefs'][band][2]
                if b1_idx is not None:
                    features[b1_idx, day_idx] = cold_curve['coefs'][band][3]
                if rmse_idx is not None:
                    features[rmse_idx, day_idx] = cold_curve['rmse'][band]
        # print(features)
                    # In this case, we didn't find a matching cold
                    # curve for the current ordinal day, stop
                    # processing.
                    # raise NoMatchingColdCurve
    except NoMatchingColdCurve:
        ...

    return features


if __name__ == '__main__':
    export_cold_main()



"""
Fork of python-slugify.

https://pypi.org/project/python-slugify/1.2.2/
"""
import re
import unicodedata
# import types
import sys

# try:
#     from htmlentitydefs import name2codepoint
#     _unicode = unicode  # NOQA
#     _unicode_type = types.UnicodeType
# except ImportError:
from html.entities import name2codepoint
_unicode = str
_unicode_type = str
unichr = chr

try:
    import text_unidecode as unidecode
except ImportError:
    import unidecode

__all__ = ['slugify', 'smart_truncate']


CHAR_ENTITY_PATTERN = re.compile(r'&(%s);' % '|'.join(name2codepoint))
DECIMAL_PATTERN = re.compile(r'&#(\d+);')
HEX_PATTERN = re.compile(r'&#x([\da-fA-F]+);')
QUOTE_PATTERN = re.compile(r'[\']+')
ALLOWED_CHARS_PATTERN = re.compile(r'[^-a-z0-9]+')
ALLOWED_CHARS_PATTERN_WITH_UPPERCASE = re.compile(r'[^-a-zA-Z0-9]+')
DUPLICATE_DASH_PATTERN = re.compile(r'-{2,}')
NUMBERS_PATTERN = re.compile(r'(?<=\d),(?=\d)')
DEFAULT_SEPARATOR = '-'


def _trunc_op(string, max_length, trunc_loc, hash_len=None, head='~', tail='~'):
    """
    max_length

    string = 'DarnOvercastSculptureTipperBlazerConcaveUnsuitedDerangedHexagonRockband'
    max_length = 16
    trunc_loc = 0.5
    _trunc_op(string, max_length, trunc_loc)
    """
    total_len = len(string)
    mid_pos = int(total_len * trunc_loc)

    num_remove = max(total_len - max_length, 1)
    import ubelt as ub
    import numpy as np
    if hash_len is None:
        recommend = min(max(4, int(np.ceil(np.log(num_remove)))), 32)
        hash_len = min(max_length, min(num_remove, recommend))

    num_insert = hash_len + len(head) + len(tail)

    actual_remove = num_remove + num_insert

    low_pos = max(0, (mid_pos - (actual_remove) // 2))
    high_pos = min(total_len, (mid_pos + (actual_remove) // 2))
    if low_pos <= 0:
        n_extra = actual_remove - (high_pos - low_pos)
        high_pos += n_extra
    if high_pos >= total_len:
        n_extra = actual_remove - (high_pos - low_pos)
        low_pos -= n_extra

    really_removed = (high_pos - low_pos)
    high_pos += (really_removed - actual_remove)

    begin = string[:low_pos]
    mid = string[low_pos:high_pos]
    end = string[high_pos:]

    mid = ub.hash_data(string)[0:hash_len]
    trunc_text = ''.join([begin, head, mid, tail, end])
    return trunc_text


def smart_truncate(string, max_length=0, word_boundary=False, separator=' ', save_order=False, trunc_loc=0.5, hash_len=None, head='~', tail='~'):
    """
    Truncate a string.
    :param string (str): string for modification
    :param max_length (int): output string length
    :param word_boundary (bool):
    :param save_order (bool): if True then word order of output string is like input string
    :param separator (str): separator between words
    :param trunc_loc (float): fraction of location where to remove the text
    :return:
    """

    string = string.strip(separator)

    if not max_length:
        return string

    if len(string) < max_length:
        return string

    if not word_boundary:
        return _trunc_op(string, max_length, trunc_loc, hash_len=hash_len, head=head, tail=tail).strip(separator)

    if separator not in string:
        return _trunc_op(string, max_length, trunc_loc, hash_len=hash_len, head=head, tail=tail)

    # hack
    truncated = ''
    # for word in string.split(separator):
    #     if word:
    #         next_len = len(truncated) + len(word)
    #         if next_len < max_length:
    #             truncated += '{}{}'.format(word, separator)
    #         elif next_len == max_length:
    #             truncated += '{}'.format(word)
    #             break
    #         else:
    #             if save_order:
    #                 break

    if not truncated:  # pragma: no cover
        truncated = _trunc_op(string, max_length, trunc_loc, hash_len=hash_len, head=head, tail=tail)
    return truncated.strip(separator)


def slugify(text, entities=True, decimal=True, hexadecimal=True, max_length=0, word_boundary=False,
            separator=DEFAULT_SEPARATOR, save_order=False, stopwords=(), regex_pattern=None, lowercase=True,
            replacements=(), trunc_loc=1.0):
    """
    Make a slug from the given text.
    :param text (str): initial text
    :param entities (bool): converts html entities to unicode
    :param decimal (bool): converts html decimal to unicode
    :param hexadecimal (bool): converts html hexadecimal to unicode
    :param max_length (int): output string length
    :param word_boundary (bool): truncates to complete word even if length ends up shorter than max_length
    :param save_order (bool): if parameter is True and max_length > 0 return whole words in the initial order
    :param separator (str): separator between words
    :param stopwords (iterable): words to discount
    :param regex_pattern (str): regex pattern for allowed characters
    :param lowercase (bool): activate case sensitivity by setting it to False
    :param replacements (iterable): list of replacement rules e.g. [['|', 'or'], ['%', 'percent']]
    :return (str):

    # Example:
    #     >>> from watch.utils.slugify_ext import slugify  # NOQA
    #     >>> import ubelt as ub
    #     >>> text = ub.cmd('diceware -n 12')['out'].strip()
    #     >>> print('text = {!r}'.format(text))
    #     >>> slug = slugify(text, max_length=10, lowercase=0, trunc_loc=1.0)
    #     >>> print('slug = {!r}'.format(slug))
    #     >>> slug = slugify(text, max_length=10, lowercase=0, trunc_loc=0.8)
    #     >>> print('slug = {!r}'.format(slug))
    #     >>> slug = slugify(text, max_length=10, lowercase=0, trunc_loc=0.5)
    #     >>> print('slug = {!r}'.format(slug))
    #     >>> slug = slugify(text, max_length=10, lowercase=0, trunc_loc=0.2)
    #     >>> print('slug = {!r}'.format(slug))
    #     >>> slug = slugify(text, max_length=10, lowercase=0, trunc_loc=0.0)
    #     >>> print('slug = {!r}'.format(slug))
    """

    # user-specific replacements
    if replacements:
        for old, new in replacements:
            text = text.replace(old, new)

    # ensure text is unicode
    if not isinstance(text, _unicode_type):
        text = _unicode(text, 'utf-8', 'ignore')

    # replace quotes with dashes - pre-process
    text = QUOTE_PATTERN.sub(DEFAULT_SEPARATOR, text)

    # decode unicode
    text = unidecode.unidecode(text)

    # ensure text is still in unicode
    if not isinstance(text, _unicode_type):
        text = _unicode(text, 'utf-8', 'ignore')

    # character entity reference
    if entities:
        text = CHAR_ENTITY_PATTERN.sub(lambda m: unichr(name2codepoint[m.group(1)]), text)

    # decimal character reference
    if decimal:
        try:
            text = DECIMAL_PATTERN.sub(lambda m: unichr(int(m.group(1))), text)
        except Exception:
            pass

    # hexadecimal character reference
    if hexadecimal:
        try:
            text = HEX_PATTERN.sub(lambda m: unichr(int(m.group(1), 16)), text)
        except Exception:
            pass

    # translate
    text = unicodedata.normalize('NFKD', text)
    if sys.version_info < (3,):
        text = text.encode('ascii', 'ignore')

    # make the text lowercase (optional)
    if lowercase:
        text = text.lower()

    # remove generated quotes -- post-process
    text = QUOTE_PATTERN.sub('', text)

    # cleanup numbers
    text = NUMBERS_PATTERN.sub('', text)

    # replace all other unwanted characters
    if lowercase:
        pattern = regex_pattern or ALLOWED_CHARS_PATTERN
    else:
        pattern = regex_pattern or ALLOWED_CHARS_PATTERN_WITH_UPPERCASE
    text = re.sub(pattern, DEFAULT_SEPARATOR, text)

    # remove redundant
    text = DUPLICATE_DASH_PATTERN.sub(DEFAULT_SEPARATOR, text).strip(DEFAULT_SEPARATOR)

    # remove stopwords
    if stopwords:
        if lowercase:
            stopwords_lower = [s.lower() for s in stopwords]
            words = [w for w in text.split(DEFAULT_SEPARATOR) if w not in stopwords_lower]
        else:
            words = [w for w in text.split(DEFAULT_SEPARATOR) if w not in stopwords]
        text = DEFAULT_SEPARATOR.join(words)

    # finalize user-specific replacements
    if replacements:
        for old, new in replacements:
            text = text.replace(old, new)

    # smart truncate if requested
    if max_length > 0:
        text = smart_truncate(text, max_length, word_boundary, DEFAULT_SEPARATOR, save_order, trunc_loc=trunc_loc)

    if separator != DEFAULT_SEPARATOR:
        text = text.replace(DEFAULT_SEPARATOR, separator)

    return text



import os
import random

import cv2
import numpy as np
from scipy import stats

import torch
from torch.utils.data import Dataset


class BaseDataset(Dataset):
    def __init__(
        self,
        dset_dir,
        split,
        video_slice,
        task_mode,
        transforms=None,
        seed_num=0,
        ignore_index=-1,
        normalize_mode=None,
        channels="RGB",
        max_iterations=None,
    ):
        super().__init__()

        self.split = split
        self.channels = channels
        self.dset_dir = dset_dir
        self.task_mode = task_mode
        self.transforms = transforms
        self.scale = video_slice.scale
        self.video_slice = video_slice
        self.ignore_index = ignore_index
        self.n_frames = video_slice.n_frames
        self.normalize_mode = normalize_mode
        self.max_iterations = max_iterations

        # Set random seed if not equal to -1.
        self.set_seed(seed_num)

        # Get global normalization parameters.
        if self.normalize_mode == "global":
            self.mean_vals, self.std_vals = self.get_pixel_normalization_params()

    def __len__(self):
        if (self.max_iterations is not None) and (self.max_iterations < len(self.examples)):
            return self.max_iterations
        else:
            return len(self.examples)

    def __getitem__(self, index):
        task_mode_fns = {
            "total_bin_change": self.get_total_binary_change_example,
            "total_sem_change": self.get_total_semantic_change_example,
            "pw_bin_change": self.get_pairwise_binary_change_example,
            "pw_sem_change": self.get_pairwise_semantic_change_example,
            "future_frame_pred": self.get_future_frame_prediction_example,
            "sem_seg": self.get_semantic_segmentation_example,
            "ss_triplet": self.get_self_supervised_triplet_example,
            "ss_splice_change": self.get_self_supervised_splice_change_example,
            "ss_crop_splice_change": self.get_self_supervised_crop_splice_change_example,
            "ss_arrow_of_time": self.get_self_supervised_arrow_of_time_example,
            "ss_mat_change": self.get_self_supervised_material_change_example,
            "ss_mat_recon": self.get_self_supervised_material_reconstruction_example,
            "ss_splice_change_index": self.get_self_supervised_splice_change_index_example,
            "bas": self.get_broad_area_search_example,
            "refine_sc": self.get_refine_site_characterization_example,
        }
        try:
            task_mode_fn = task_mode_fns[self.task_mode]
        except KeyError:
            raise NotImplementedError(f'Target task "{self.task_mode}" has not been implemented in BaseDataset class.')
        example = task_mode_fn(index)
        return example

    def set_seed(self, seed_num):
        if seed_num == -1:
            print("No random seed set!")
        else:
            print(f"Setting random seed to: {seed_num}")
            torch.backends.cudnn.enabled = True
            torch.backends.cudnn.benchmark = True
            torch.backends.cudnn.deterministic = False
            torch.manual_seed(seed_num)
            torch.cuda.manual_seed_all(seed_num)
            np.random.seed(seed_num)
            random.seed(seed_num)

    def get_total_binary_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_total_semantic_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_pairwise_binary_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_pairwise_semantic_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_future_frame_prediction_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_semantic_segmentation_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_triplet_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_splice_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_crop_splice_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_arrow_of_time_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_material_change_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_splice_change_index_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_self_supervised_material_reconstruction_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_refine_site_characterization_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def get_broad_area_search_example(self, index):
        raise NotImplementedError(f'Task "{self.task_mode}" not implemented for this dataset.')

    def to_rgb(self, image):
        raise NotImplementedError("To RGB method not implemented for this dataset.")

    def load_frames(self, image_paths, crop_slice):
        raise NotImplementedError("Load frames not implemented for this dataset.")

    def colorize_target_mask(self, target_mask):
        raise NotImplementedError("Load frames not implemented for this dataset.")

    def compute_class_distribution(self, target_mask, n_classes, ignore_index=-1):
        """Compute the class distribution of a target image as percentages.

        Args:
            target_mask (_type_): TODO: _description_
            n_classes (_type_): TODO: _description_
            ignore_index (int, optional): TODO: _description_. Defaults to -1.
        """
        unique_classes = list(np.unique(target_mask))

        # Remove the ignore index class.
        if ignore_index in unique_classes:
            unique_classes.remove(ignore_index)

        class_dist = np.zeros(n_classes)
        for class_number in unique_classes:
            x, y = np.where(target_mask == class_number)
            class_dist[class_number] = x.shape[0]

        # Normalize the class distribution to a probability.
        class_dist /= class_dist.sum()

        return class_dist

    def visualize_example(self, index, save_path=None, num_plot_images=5, overlay_last_anno=False):
        import matplotlib.pyplot as plt

        # Get an example.
        example = self.__getitem__(index)

        if self.task_mode is None:
            # Assumes that there are no labels in the example.
            F = int(example["active_frames"].sum())

            num_plot_images += 1

            if F >= num_plot_images:
                # Fewer images in sequence than number of plot images.
                fig, axes = plt.subplots(1, num_plot_images, figsize=(30, 5))
            else:
                # More images in sequence than number of plot images.
                fig, axes = plt.subplots(1, F, figsize=(30, 5))

            if F >= num_plot_images:
                # More images in sequence than number of plot images.
                indices = [(F // num_plot_images) * i for i in range(num_plot_images - 1)]
                indices.append(F - 1)

                for i, index in enumerate(indices):
                    rgb_image = self.to_rgb(example["video"][index])

                    axes[i].imshow(rgb_image)
                    axes[i].set_title(f'{example["datetimes"][i].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i].axis("off")
            else:
                # Fewer images in sequence than number of plot images.
                for i in range(num_plot_images):
                    rgb_image = self.to_rgb(example["video"][i])  # TODO:

                    axes[i].imshow(rgb_image)
                    axes[i].set_title(f'{example["datetimes"][i].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i].axis("off")

            if "region_name" in list(example.keys()):
                fig.suptitle(example["region_name"])

            if save_path is not None:
                if overlay_last_anno:
                    save_path = os.path.splitext(save_path)[0] + "_overlay" + os.path.splitext(save_path)[1]
                print(f"Saving figure to path: {save_path}")
                plt.savefig(save_path, dpi=300)

        elif self.task_mode == "ss_arrow_of_time":
            # Assumes that there are no labels in the example.
            F = int(example["active_frames"].sum())

            reverse = example[self.task_mode]  # [0 - not reversed, 1 - reversed]
            reverse = reverse == 1

            num_plot_images += 1

            if F >= num_plot_images:
                # Fewer images in sequence than number of plot images.
                fig, axes = plt.subplots(1, num_plot_images, figsize=(30, 5))
            else:
                # More images in sequence than number of plot images.
                fig, axes = plt.subplots(1, F, figsize=(30, 5))

            if F >= num_plot_images:
                # More images in sequence than number of plot images.
                indices = [(F // num_plot_images) * i for i in range(num_plot_images - 1)]
                indices.append(F - 1)

                for i, index in enumerate(indices):
                    rgb_image = self.to_rgb(
                        self.unnormalize(example["video"][index])[0],
                        mean=example["mean"][index],
                        std=example["std"][index],
                    )

                    axes[i].imshow(rgb_image)
                    axes[i].set_title(f'{example["datetimes"][index].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i].axis("off")
            else:
                # Fewer images in sequence than number of plot images.
                for i in range(num_plot_images):
                    rgb_image = self.to_rgb(
                        self.unnormalize(example["video"][i])[0], mean=example["mean"][i], std=example["std"][i]
                    )

                    axes[i].imshow(rgb_image)
                    axes[i].set_title(f'{example["datetimes"][i].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i].axis("off")

            # Add title to entire resquence.
            if "region_name" in list(example.keys()):
                fig.suptitle(f'{example["region_name"]} | {self.task_mode} | Reverse: {reverse}')
            else:
                fig.suptitle(f"{self.task_mode} | Reverse: {reverse}")

            if save_path is not None:
                if overlay_last_anno:
                    save_path = os.path.splitext(save_path)[0] + "_overlay" + os.path.splitext(save_path)[1]
                print(f"Saving figure to path: {save_path}")
                plt.savefig(save_path, dpi=300)

        elif self.task_mode == "total_bin_change":
            # Plot a subset of images in the sequence and change mask
            F = int(example["active_frames"].sum())

            if F >= num_plot_images:
                # Fewer images in sequence than number of plot images.
                fig, axes = plt.subplots(1, num_plot_images + 1, figsize=(30, 5))
            else:
                # More images in sequence than number of plot images.
                fig, axes = plt.subplots(1, F + 1, figsize=(30, 5))

            # Plot change mask
            label_image = np.clip(example[self.task_mode], 0, 100)
            axes[0].imshow(label_image.float(), cmap="gray")
            axes[0].set_title("Sequence Binary Change Mask")
            axes[0].axis("off")

            if F >= num_plot_images:
                # More images in sequence than number of plot images.
                indices = [(F // num_plot_images) * i for i in range(num_plot_images - 1)]
                indices.append(F - 1)

                for i, index in enumerate(indices):
                    rgb_image = self.to_rgb(
                        self.unnormalize(example["video"][index], mean=example["mean"], std=example["std"])
                    )

                    if overlay_last_anno and i == (len(indices) - 1):
                        # Format label image
                        mpl_label_image = np.ma.masked_where(label_image.numpy() == 0, label_image.numpy())
                        axes[i + 1].imshow(rgb_image, interpolation="none")
                        axes[i + 1].imshow(mpl_label_image, interpolation="none", alpha=0.4, cmap="cool")
                    else:
                        axes[i + 1].imshow(rgb_image)
                    axes[i + 1].set_title(f'{example["datetimes"][index].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i + 1].axis("off")
            else:
                # Fewer images in sequence than number of plot images.
                for i in range(num_plot_images):
                    rgb_image = self.to_rgb(
                        self.unnormalize(example["video"][i], mean=example["mean"], std=example["std"])
                    )
                    if overlay_last_anno and i == (num_plot_images - 1):
                        # Format label image
                        mpl_label_image = np.ma.masked_where(label_image.numpy() == 0, label_image.numpy())
                        axes[i + 1].imshow(rgb_image, interpolation="none")
                        axes[i + 1].imshow(mpl_label_image, interpolation="none", alpha=0.4, cmap="cool")
                    else:
                        axes[i + 1].imshow(rgb_image)

                    axes[i + 1].set_title(f'{example["datetimes"][i].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i + 1].axis("off")

            if "region_name" in list(example.keys()):
                fig.suptitle(example["region_name"])

            if save_path is not None:
                if overlay_last_anno:
                    save_path = os.path.splitext(save_path)[0] + "_overlay" + os.path.splitext(save_path)[1]
                print(f"Saving figure to path: {save_path}")
                plt.savefig(save_path, dpi=300)

        elif self.task_mode == "sem_seg":
            assert not overlay_last_anno, "Overlaying annotation for sem_seg is not implmented yet."
            # Plot a subset of images in the sequence and change mask
            F = int(example["active_frames"].sum())

            if F >= num_plot_images:
                # Fewer images in sequence than number of plot images.
                fig, axes = plt.subplots(1, num_plot_images + 1, figsize=(30, 5))
            else:
                # More images in sequence than number of plot images.
                fig, axes = plt.subplots(1, F + 1, figsize=(30, 5))

            # Plot change mask
            label_image = example[self.task_mode]
            color_label_mask = self.colorize_target_mask(label_image)
            axes[0].imshow(color_label_mask)
            axes[0].set_title("Sequence Semantic Change Mask")
            axes[0].axis("off")

            if F >= num_plot_images:
                # More images in sequence than number of plot images.
                indices = [(F // num_plot_images) * i for i in range(num_plot_images - 1)]
                indices.append(F - 1)

                for i, index in enumerate(indices):
                    rgb_image = self.to_rgb(example["video"][index])

                    axes[i + 1].imshow(rgb_image)
                    axes[i + 1].set_title(f'{example["datetimes"][index].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i + 1].axis("off")
            else:
                # Fewer images in sequence than number of plot images.
                for i in range(num_plot_images):
                    rgb_image = self.to_rgb(example["video"][i])

                    axes[i + 1].imshow(rgb_image)
                    axes[i + 1].set_title(f'{example["datetimes"][i].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                    axes[i + 1].axis("off")

            if "region_name" in list(example.keys()):
                fig.suptitle(example["region_name"])

            if save_path is not None:
                print(f"Saving figure to path: {save_path}")
                plt.savefig(save_path, dpi=300)

        elif self.task_mode == "ss_splice_change_index":
            # Get change index.
            change_index = example[self.task_mode].item()

            # Plot a subset of images in the sequence and change mask
            F = int(example["active_frames"].sum())

            if F <= num_plot_images:
                num_plot_images = F
                indices = list(range(num_plot_images))
            else:
                ## Take a subset of frames but include the change frame.
                indices = [(F // num_plot_images) * i for i in range(num_plot_images - 2)]
                indices.append(F - 1)
                indices.append(change_index)
                indices = sorted(indices)

            fig, axes = plt.subplots(1, num_plot_images, figsize=(30, 5))

            for i, index in enumerate(indices):
                rgb_image = self.to_rgb(
                    self.unnormalize(example["video"][index], mean=example["mean"], std=example["std"])
                )

                axes[i].imshow(rgb_image)
                if i == change_index:
                    # Format label image
                    axes[i].set_title(
                        f'{example["datetimes"][index].strftime("%m/%d/%Y")} | [{index}/{F-1}]', color="red"
                    )
                else:
                    axes[i].set_title(f'{example["datetimes"][index].strftime("%m/%d/%Y")} | [{index}/{F-1}]')
                axes[i].axis("off")

            if "region_name" in list(example.keys()):
                fig.suptitle(example["region_name"])

            if save_path is not None:
                print(f"Saving figure to path: {save_path}")
                plt.savefig(save_path, dpi=300)
        elif self.task_mode == "segmentation":
            # A row of RGB images
            # A row of segmentation images
            raise NotImplementedError(self.task_mode)
        else:
            raise NotImplementedError(self.task_mode)
        plt.close()

    def generate_visualization_GIF(self, index, save_path, overlay_annos=False, overlay_dates=True):
        import cv2
        from PIL import Image
        import matplotlib.pyplot as plt
        from transformer.utils.util_misc import create_gif

        save_dir = os.path.split(save_path)[0]
        save_name = os.path.splitext(os.path.split(save_path)[1])[0]
        save_ext = os.path.splitext(save_path)[1]

        if (self.task_mode is None) or (self.task_mode == "ss_mat_recon"):
            example = self.__getitem__(index)

            # Get all activate frames in video
            F = int(example["active_frames"].sum())
            rgb_frames = []
            for i in range(F):
                # Load frame and convert to RGB.
                rgb_frame = self.to_rgb(example["video"][i])

                rgb_frames.append(rgb_frame)

            # Add date and frame number to images.
            if overlay_dates:
                img_text = [
                    f"[{i+1}/{len(example['datetimes'])}]: " + dt.strftime("%m-%d-%Y")
                    for i, dt in enumerate(example["datetimes"])
                ]
            else:
                img_text = None

            # Create a GIF of RGB frames.
            create_gif(rgb_frames, save_path, fps=1, image_text=img_text)

        elif self.task_mode == "total_bin_change":
            example = self.__getitem__(index)

            # Get all activate frames in video
            F = int(example["active_frames"].sum())
            rgb_frames = []
            for i in range(F):
                # Load frame and convert to RGB.
                rgb_frame = self.to_rgb(self.unnormalize(example["video"][i], mean=example["mean"], std=example["std"]))

                rgb_frames.append(rgb_frame)

            # Add date and frame number to images.
            if overlay_dates:
                img_text = [
                    f"[{i+1}/{len(example['datetimes'])}]: " + dt.strftime("%m-%d-%Y")
                    for i, dt in enumerate(example["datetimes"])
                ]
            else:
                img_text = None

            # Save ground truth image separately.
            target = (np.clip(example[self.task_mode].numpy(), 0, 1) * 255).astype("uint8")
            target_path = os.path.join(save_dir, save_name + "_gt.png")
            Image.fromarray(target).save(target_path)

            if overlay_annos:
                # Convert the annotation to edges.
                contours, _ = cv2.findContours(target, mode=cv2.RETR_EXTERNAL, method=cv2.CHAIN_APPROX_NONE)

                # Create plot of contours.
                fig, ax = plt.subplots(1, 1)
                canvas = np.zeros(target.shape)
                plt.imshow(canvas)
                for poly in contours:
                    # Close the loop of the contour.
                    poly = np.concatenate((poly, poly[0][np.newaxis]), axis=0)

                    # Get x and y values.
                    xs, ys = poly[..., 0][:, 0], poly[..., 1][:, 0]

                    # Draw polygon on canvas.
                    plt.plot(xs, ys, "r-", linewidth=0.1)
                contour_save_path = os.path.join(save_dir, save_name + "_contour.png")
                plt.savefig(contour_save_path, dpi=300)
                plt.close()

                overlayed_rgb_frames = []
                for rgb_frame in rgb_frames:
                    H, W, _ = rgb_frame.shape

                    # Try using matplotlib to overlay polygons on RGB image.

                    # Make content fill the entire figure.
                    plt.figure(frameon=False)
                    plt.imshow(rgb_frame)
                    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
                    ax.set_axis_off()
                    fig.add_axes(ax)

                    # ax.imshow(rgb_frame, aspect='auto')
                    ax.imshow(rgb_frame)

                    for poly in contours:
                        # Close the loop of the contour.
                        poly = np.concatenate((poly, poly[0][np.newaxis]), axis=0)

                        # Get x and y values.
                        xs, ys = poly[..., 0][:, 0], poly[..., 1][:, 0]

                        # Draw polygon on canvas.
                        ax.plot(xs, ys, "r-", linewidth=0.5)

                    # Convert figure to rgb numpy array.
                    fig.canvas.draw()
                    image = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
                    image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))

                    overlayed_rgb_frames.append(image)
                    plt.close()

                # Create a GIF.
                save_path = os.path.join(save_dir, save_name + "_overlay" + save_ext)
                create_gif(overlayed_rgb_frames, save_path, fps=1, image_text=img_text)
            else:
                # Create a GIF of RGB frames.
                create_gif(rgb_frames, save_path, fps=1, image_text=img_text)

        else:
            raise NotImplementedError(f"GIF generation for mode {self.task_mode} not implemented.")

    def get_subset_frame_ids(self, n_video_frames, n_req_frames):
        assert n_video_frames > 1

        if n_video_frames == 2:
            return [0, 1]
        elif n_video_frames == n_req_frames:
            return len(range(n_video_frames))
        else:
            # Divide in half N-2 times and find the middle index.
            N = n_video_frames
            M = n_req_frames

            middle_index = N // 2
            frame_ids = [middle_index]
            while len(frame_ids) != (M - 2):
                if len(frame_ids) % 2 == 1:
                    # Get the right middle half.
                    right_half_index = middle_index + frame_ids[-1] // 2
                    frame_ids.append(right_half_index)
                else:
                    # Get the left middle half.
                    # Get the right middle half.
                    left_half_index = middle_index - frame_ids[-2] // 2
                    frame_ids.append(left_half_index)
        frame_ids.insert(0, 0)
        frame_ids.append(n_video_frames - 1)
        frame_ids = sorted(frame_ids)
        return frame_ids

    def visualize_dates(self, index, save_path=None):
        import matplotlib.pyplot as plt
        from collections import defaultdict

        # Get an example.
        example = self.__getitem__(index)

        if "datetimes" not in example.keys():
            raise KeyError("Dataset does not contain datetimes in example.")

        # Get unique years in datetimes.
        dates = example["datetimes"]
        years = np.asarray([date.year for date in dates])
        years = list(np.unique(years))

        # Spilt datetimes.
        split_dict = defaultdict(list)
        for date in dates:
            yday = date.timetuple().tm_yday
            split_dict[date.year].append(yday)

        # Plot datetimes.
        _, axes = plt.subplots(len(years), 1)

        for i, (year, ydays) in enumerate(split_dict.items()):
            canvas = np.zeros([5, 365], dtype=float)
            for yday in ydays:
                canvas[:, yday] = 1.0
            axes[i].imshow(canvas, cmap="Blues")
            axes[i].set_ylabel(year)
            axes[i].tick_params(left=False, bottom=False, labelleft=False, labelbottom=True)
        plt.tight_layout()

        if save_path is not None:
            print(f"Saving figure to path: {save_path}")
            plt.savefig(save_path, dpi=300)

    def scale_video(self, video, scale_factor, inter_mode=cv2.INTER_NEAREST):
        """Up or down scale all frames of video.

        Args:
            video (numpy array): A float numpy array of shape [n_frames, n_channels, height, width].
            scale_factor (int/float): The factor to scale the video and target by. NOTE: This variable be between (0, inf).
            inter_mode (int, optional): An OpenCV resize parameter to determine the resizing method. Defaults to cv2.INTER_NEAREST.

        Returns:
            numpy array: A resized video.
        """
        # Rearrange axes from [n_frames, n_channels, height, width] to [n_frames, height, width, n_channels].
        video = video.transpose(0, 2, 3, 1)

        # Resize video.
        rz_video = []
        for frame_index in range(video.shape[0]):
            rz_frame = self.scale_frame(video[frame_index], scale_factor, inter_mode=inter_mode)
            rz_video.append(rz_frame)
        video = np.stack(rz_video, axis=0)

        # Rearrange axes from [n_frames, height, width, n_channels] to [n_frames, n_channels, height, width].
        video = video.transpose(0, 3, 1, 2)

        return video

    def scale_frame(self, frame, scale_factor, inter_mode=cv2.INTER_NEAREST):
        """Up or down scale a frame.

        Args:
            frame (numpy array): A numpy array of shape [n_channels, height, width].
            scale_factor (int/float): The factor to scale the video and target by. NOTE: This variable be between (0, inf).
            inter_mode (int, optional): An OpenCV resize parameter to determine the resizing method. Defaults to cv2.INTER_NEAREST.

        Returns:
            numpy array: A resized frame.
        """
        rz_frame = cv2.resize(
            frame,
            (int(self.video_slice.height * scale_factor), int(self.video_slice.width * scale_factor)),
            interpolation=inter_mode,
        )
        return rz_frame

    def get_pixel_normalization_params(self):
        raise NotImplementedError

    def normalize(self, video):
        """Normalize videos before they are transformed.

        Args:
            video (torch.tensor): Shape [n_frames, channels, height, width]

        Returns:
            torch.tensor: Shape [n_frames, channels, height, width]
        """
        if self.normalize_mode is None:
            mean, std = [0], [1]
        elif self.normalize_mode == "global":
            mean, std = self.mean_vals, self.std_vals
            video = (video - self.mean_vals) / self.std_vals
        elif self.normalize_mode == "local":
            # Compute the video mean and std.
            mean = np.mean(video, axis=(0, 2, 3))
            std = np.std(video, axis=(0, 2, 3))

            if sum(mean) == 0:
                mean = np.zeros(mean.shape[0])
            mean = np.nan_to_num(mean)

            if sum(std) == 0:
                std = np.ones(std.shape[0])

            indices = np.where(std == 0)
            std[indices] = 1

            # Format mean and std shapes.
            mean = mean[np.newaxis, :, np.newaxis, np.newaxis]
            std = std[np.newaxis, :, np.newaxis, np.newaxis]

            self.local_mean, self.local_std = mean, std
            video = (video - self.local_mean) / self.local_std

        elif self.normalize_mode == "local_trim":
            # Compute the trimmed video mean and std over channels.
            # Trim the top and bottom 15%.
            trimmed_pct = 0.15
            channel_means, channel_stds = [], []
            for channel_index in range(video.shape[1]):
                # Compute trimmed mean.
                try:
                    channel_mean = stats.trim_mean(video[:, channel_index], trimmed_pct, axis=None)
                except Exception:
                    print("Upset trim mean")
                    channel_mean = 0
                channel_means.append(channel_mean)

                # Compute trimmed mean.
                try:
                    channel_std = stats.mstats.trimmed_std(video[:, channel_index], trimmed_pct, axis=None)
                except Exception:
                    print("Upset trim STD")
                    channel_std = 1
                channel_stds.append(channel_std)

            mean = np.asarray(channel_means)
            std = np.asarray(channel_stds)

            if sum(mean) == 0:
                mean = np.zeros(mean.shape[0])
            mean = np.nan_to_num(mean)

            if sum(std) == 0:
                std = np.ones(std.shape[0])

            indices = np.where(std == 0)
            std[indices] = 1

            # Format mean and std shapes.
            mean = mean[np.newaxis, :, np.newaxis, np.newaxis]
            std = std[np.newaxis, :, np.newaxis, np.newaxis]

            self.local_trim_mean, self.local_trim_std = mean, std

            video = (video - self.local_trim_mean) / self.local_trim_std

        else:
            raise NotImplementedError(f"Normalize mode: {self.normalize_mode}")

        mean, std = torch.tensor(mean), torch.tensor(std)
        return video, mean, std

    def unnormalize(self, video, mean=None, std=None):
        """Undo the normalization process.

        Args:
            video (torch.tensor): Shape [n_frames, channels, height, width]

        Returns:
            torch.tensor: Shape [n_frames, channels, height, width]
        """
        if mean is not None:
            video = (video * std) + mean
            return video

        if self.normalize_mode is None:
            video = video[None]
        elif self.normalize_mode == "global":
            video = (video * self.std_vals) + self.mean_vals
        elif self.normalize_mode == "local":
            video = (video * self.local_std) + self.local_mean
        elif self.normalize_mode == "local_trim":
            video = (video * self.local_trim_std) + self.local_trim_mean
        else:
            raise NotImplementedError(f"Normalize mode: {self.normalize_mode}")

        return video

    def pad_video_height_width(self, video, target_height, target_width):
        """Pad a video to the bottom and right of the frames if not same resolution as target height and width.

        Args:
            video (np.array): A numpy array of shape [n_frames, n_channels, og_height, og_width].
            target_height (int): The target height resolution parameter.
            target_width (int): The target width resolution parameter.

        Returns:
            [np.array]: A numpy array of shape [n_frames, n_channels, target_height, target_width].
        """

        n_frames, n_channels, og_height, og_width = video.shape

        if (og_height < target_height) or (og_width < target_width):
            canvas = np.ones((n_frames, n_channels, target_height, target_width), dtype=video.dtype) * self.pad_value
            canvas[:, :, :og_height, :og_width] = video
            video = canvas
        elif (og_height > target_height) or (og_width > target_width):
            raise ValueError(
                f"Input video of resolution [{og_height},{og_width}] is larger than target shape [{target_height},{target_width}]."
            )

        return video



"""
moved to watch.utils.util_globals
"""
from watch.utils.util_eval import restricted_eval, RestrictedSyntaxError  # NOQA
from watch.utils.util_resources import request_nofile_limits, check_shm_limits  # NOQA
from watch.utils.util_globals import coerce_num_workers, configure_global_attributes  # NOQA



# from models.unet_stn import  UNet
from watch.tasks.rutgers_material_seg.models import deeplabWS
from watch.tasks.rutgers_material_seg.models.shallow_seg import ShallowSeg
from watch.tasks.rutgers_material_seg.models.linear_classifier import VSNet
from watch.tasks.rutgers_material_seg.models import resnet
from watch.tasks.rutgers_material_seg.models import resnetGNWS
from watch.tasks.rutgers_material_seg.models import resnet_enc
from watch.tasks.rutgers_material_seg.models import deeplab
from watch.tasks.rutgers_material_seg.models import deeplab_diff
from watch.tasks.rutgers_material_seg.models import resnet_classification_finetune

models = {
    'deeplabWS': deeplabWS,
    'shallow_seg': ShallowSeg,
    'vsnet': VSNet,
    'resnet': resnet,
    'resnetgnws': resnetGNWS,
    'resnet_enc': resnet_enc,
    'deeplab': deeplab,
    'resnet_class_ft': resnet_classification_finetune,
    "deeplab_diff": deeplab_diff
}


def build_model(model_name: str = "deeplabWS", backbone: str = "resnet101", **kwargs) -> object:
    """Model building module

    Parameters
    ----------
    model_name : str, optional
        which model to use, by default "deeplabWS"
    backbone : str, optional
        backbone of model, by default "resnet101"

    Returns
    -------
    model
        model object
    """
    if backbone:
        model = getattr(models[model_name], backbone)(**kwargs)
    else:
        model = models[model_name](**kwargs)
    return model



import torch


class VideoPatchifyzer:
    def __init__(self, frame_size, patch_size):
        """Divide and reshape a video into patches.

        Input video shape [batch_size, n_frames, n_channels, height, width] gets divded into a
        patch video of shape [batch_size, n_frames, n_channels, n_patches, patch_size, patch_size].
            Where patch_size = frame_height/patch_height.

        NOTE: The size of the frame and the patch must be squres and must be divisible.

        Args:
            frame_size (tupe(int, int)): A tuple containing each frames height and width.
            patch_size (tupe(int, int)): A tuple containing each patch height and width.
        """
        frame_height, frame_width = frame_size
        patch_height, patch_width = patch_size

        assert frame_height == frame_width, f"Frame size is not square: HW = [{frame_height},{frame_width}]"
        assert patch_height == patch_width, f"Patch size is not square: HW = [{patch_height},{patch_width}]"

        assert (
            frame_height % patch_height
        ) == 0, f"Frame size is not divisible by patch size equally: {frame_height} !/ {patch_height}"

        self.frame_size = frame_height

    def __call__(self, video):
        """Reshape the frames of a video into equal patches.

        Args:
            video (torch.tensor): A tensor of shape [batch_size, n_frames, n_channels, n_patches, patch_size, patch_size].
        """
        return video.unfold(3, 3, 3).unfold(4, 3, 3).flatten(5).permute(0, 1, 2, 5, 3, 4)


if __name__ == "__main__":
    # Test 1: Make sure that video patchifyzer is generate correct shapes.
    patchifyzer = VideoPatchifyzer(frame_size=(9, 9), patch_size=(3, 3))
    batch_size, n_frames, n_channels, height, width = 2, 5, 4, 9, 9
    test_video = torch.zeros([batch_size, n_frames, n_channels, height, width])
    target_size = torch.Size([2, 5, 4, 9, 3, 3])
    test_patch_video = patchifyzer(test_video)
    output_shape = test_patch_video.shape
    assert output_shape == target_size, f"Output patch video is shape {output_shape} but should be {target_size}"
